from bitso.media.models import MEDIA_PATH_RESOLVER
from bitso.utils import get_setting
from storages.backends.s3boto import S3BotoStorage

VERSION = get_setting("VERSION")

class S3Storage(S3BotoStorage):
    
    def url(self, name):
        
        url = S3BotoStorage.url(self, name)
        
        upload_path = None
        
        if isinstance(MEDIA_PATH_RESOLVER, (str, unicode)):
            
            upload_path = MEDIA_PATH_RESOLVER
            
        elif hasattr(MEDIA_PATH_RESOLVER, "path"):
            
            upload_path = MEDIA_PATH_RESOLVER.path
            
        if upload_path and url.__contains__(upload_path):
            
            return url
        
        if url.__contains__("?") :
        
            return url + "&version=" + VERSION
        
        return url + "?version=" + VERSION


from django.contrib import admin
from easy_thumbnails.models import Source, Thumbnail

from bitsoframework.media.settings import Image


class ImageAdmin(admin.ModelAdmin):
    model = Image
    list_display = ["id", "title", "filename", "file", "parent_type", "parent_id", "width", "height"]
    search_fields = ["title", "filename"]


class ThumbnailAdmin(admin.TabularInline):
    model = Thumbnail


class SourceAdmin(admin.ModelAdmin):
    model = Source
    list_display = ("id", "storage_hash", "modified")
    inlines = (
        ThumbnailAdmin,
    )


admin.site.register(Source, SourceAdmin)



import os
import shutil
import sys
import tempfile

from django.core.files.base import File

from bitsoframework.media.services import MediaService
from bitsoframework.media.settings import (PREVIEW_ENABLED, PREVIEW_HEIGHT,
                                           PREVIEW_WIDTH, Document, Image)
from bitsoframework.utils import files


class DocumentService(MediaService):
    """
    Service used to manage the lifecycle of the Document model.
    """

    model_class = Document

    def filter(self, **kwargs):
        queryset = super(DocumentService, self).filter(**kwargs)

        if PREVIEW_ENABLED:
            queryset = queryset.select_related("preview")

        return queryset

    def create(self, file=None, category=None, generate_preview=PREVIEW_ENABLED, **kwargs):

        document = super(DocumentService, self).create(file=file, category=category, **kwargs)

        if generate_preview and not kwargs.get("preview"):
            self.create_preview(document, file)

        return document

    def create_preview(self, document, file):
        from preview_generator.manager import PreviewManager

        preview_file_path = None
        temp_source_path = None

        try:
            temp_dir = tempfile.gettempdir()

            try:
                temp_source_path = document.file.path
            except:
                temp_source_path = tempfile.mktemp(files.get_extension(file.name))

                with open(temp_source_path, 'wb') as outfile:
                    document.file.open(mode='rb')
                    shutil.copyfileobj(document.file, outfile)

            preview_manager = PreviewManager(temp_dir, create_folder=True)

            preview_file_path = preview_manager.get_jpeg_preview(temp_source_path,
                                                                 width=PREVIEW_WIDTH,
                                                                 height=PREVIEW_HEIGHT)

            if preview_file_path:
                preview_file = File(file=None, name=preview_file_path)
                preview_file.open(mode='rb')

                preview = Image(file=preview_file)
                # temporarily store preview_source until one-to-one is loaded after save
                preview.preview_source = document
                preview.save()

                document.preview = preview
                document.old_file = document.file  # avoid re-uploading the file
                document.save()

        except Exception as e:
            import traceback
            print("Could not generate preview for file: " + str(file) + " due " + str(e))
            print(traceback.format_exception(None,  # <- type(e) by docs, but ignored
                                             e, e.__traceback__),
                  file=sys.stderr, flush=True)
        finally:
            if preview_file_path:
                os.remove(preview_file_path)



from rest_framework import exceptions, viewsets, status
from rest_framework.decorators import action
from rest_framework.exceptions import ValidationError
from rest_framework.response import Response
from rest_framework.settings import api_settings

from bitsoframework.serializers import ListIDSerializer


class CreateBulkModelMixin(object):
    """
    Create model instances in bulk.
    """

    @action(detail=False, methods=["POST"], url_path="bulk_create")
    def create_bulk(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data, many=True)
        serializer.is_valid(raise_exception=True)
        instances = self.perform_create_bulk(serializer)
        if isinstance(self, ActionSerializerView):
            serializer = self.get_serializer_for("list", instances, many=True)
        headers = self.get_success_headers(serializer.data)
        return Response(serializer.data, status=status.HTTP_201_CREATED, headers=headers)

    def perform_create_bulk(self, serializer):
        return serializer.save()

    def get_success_headers(self, data):
        try:
            return {'Location': str(data[api_settings.URL_FIELD_NAME])}
        except (TypeError, KeyError):
            return {}


class UpdateBulkModelMixin(object):
    """
    Update model instances in bulk.
    """

    @action(detail=False, methods=["PATCH"], url_path="bulk_update")
    def update_bulk(self, request, *args, **kwargs):
        ids = self.get_update_ids(request.data)
        instances = list(self.get_queryset().filter(id__in=ids))

        serializer = self.get_serializer(
            instances, data=request.data, partial=False, many=True
        )
        serializer.is_valid(raise_exception=True)

        self.perform_update_bulk(serializer)

        return Response(serializer.data)

    def perform_update_bulk(self, serializer):
        return serializer.save()

    def get_update_ids(self, data, field="id", unique=True):

        if isinstance(data, list):
            id_list = [int(x[field]) for x in data]

            if unique and len(id_list) != len(set(id_list)):
                raise ValidationError("Multiple updates to a single {} found".format(field))

            return id_list

        return [data]


class DestroyBulkModelMixin(object):
    """
    Destroy model instances in bulk.
    """

    @action(detail=False, methods=["DELETE"], url_path="bulk_destroy")
    def destroy_bulk(self, request, *args, **kwargs):
        serializer = ListIDSerializer(request.data)
        serializer.is_valid(raise_exception=True)

        self.perform_destroy_bulk(serializer)

        return Response(status=status.HTTP_204_NO_CONTENT)

    def perform_destroy_bulk(self, serializer):
        instances = self.get_queryset().filter(id__in=serializer.validated_data.get("ids"))
        instances.delete()


class MethodSerializerView(object):
    """
    Utility class for get different serializer class by http method.
    For example:
    method_serializer_classes = {
        ("GET", ): MyModelListViewSerializer,
        ("PUT", "PATCH"): MyModelCreateUpdateSerializer
    }
    """
    http_method_serializer_classes = None

    def get_serializer_class(self):
        assert self.http_method_serializer_classes is not None, (
                "Expected view %s should contain http_method_serializer_classes "
                "to get right serializer class." %
                (self.__class__.__name__,)
        )
        for methods, serializer_cls in self.http_method_serializer_classes.items():
            if self.request.method in methods:
                return serializer_cls

        raise exceptions.MethodNotAllowed(self.request.method)


class ActionSerializerView(object):
    """
    Utility class for get different serializer class by action name.
    For example:
    action_serializer_classes = {
        ("create", ): MyModelListViewSerializer,
        ("retrieve", "update"): MyModelCreateUpdateSerializer
    }
    """
    action_serializer_classes = None

    def get_serializer_for(self, action, *args, **kwargs):

        kwargs['context'] = self.get_serializer_context()

        return self.get_serializer_class_for(action)(*args, **kwargs)

    def get_serializer_class_for(self, action):
        if self.action_serializer_classes:
            for actions, serializer_cls in self.action_serializer_classes.items():
                if action in actions:
                    return serializer_cls

        return super(ActionSerializerView, self).get_serializer_class()

    def get_serializer_class(self):
        return self.get_serializer_class_for(self.action)


class SmartQuerysetView(object):

    def get_queryset(self):

        queryset = super(SmartQuerysetView, self).get_queryset()

        serializer_class = self.get_serializer_class()

        if hasattr(serializer_class, "Meta"):

            if hasattr(serializer_class.Meta, "select_related"):
                queryset = queryset.select_related(*serializer_class.Meta.select_related)

            if hasattr(serializer_class.Meta, "prefetch_related"):
                queryset = queryset.prefetch_related(*serializer_class.Meta.prefetch_related)

            if hasattr(serializer_class.Meta, "select_only"):
                queryset = queryset.only(*serializer_class.Meta.select_only)

            elif hasattr(serializer_class.Meta, "fields") and isinstance(serializer_class.Meta.fields, (list, tuple)):
                queryset = queryset.only(*serializer_class.Meta.fields)

        return queryset


class AbstractViewSet(ActionSerializerView,
                      SmartQuerysetView,
                      viewsets.GenericViewSet):
    """
    Base abstract viewset for convenience
    """
    pass



from bitsoframework.vos import AbstractVO


class MetadataModel(AbstractVO):

    def __init__(self, attributes=None, **kwargs):
        super(MetadataModel, self).__init__(**kwargs)
        self.attributes = attributes or []


class MetadataAttribute(AbstractVO):
    pass



from django.apps import apps

from bitsoframework.metadata.models import MetadataModel


class MetadataModelService(object):

    def __init__(self):
        self.models = {}

    def get_model(self, app_label, model_name):
        name = "%s.%s" % (app_label, model_name)
        if name not in self.models:
            self.models[name] = self.build_model(app_label, model_name)

        return self.models.get(name)

    def build_model(self, app_label, model_name):
        django_model = apps.get_model(app_label, model_name)
        model = MetadataModel(name=model_name)

        return model



from django.db.models.base import Model

from bitsoframework.utils.models import set_attrs


class CopyableModel(Model):
    class Meta:
        abstract = True

    def copy(self, auto_save=True, **kwargs):
        pk = self.pk

        copy = self.__class__.objects.get(id=pk)
        copy.pk = None
        copy.id = None

        set_attrs(self, kwargs)

        if auto_save:
            copy.save(force_insert=True)

        return copy



from django.contrib.gis.db.models import PointField
from django.db import models
from django.db.models import Model


class AbstractAddress(Model):
    address = models.CharField(max_length=250, null=True, blank=True)
    city = models.CharField(max_length=250, db_index=True, null=True, blank=True)
    state = models.CharField(max_length=250, db_index=True, null=True, blank=True)
    country = models.CharField(max_length=250, db_index=True, null=True, blank=True)
    postal_code = models.CharField(max_length=7, null=True, blank=True)

    full_address = models.CharField(max_length=1200, null=True, blank=True)

    coordinates = PointField(null=True, blank=True)

    class Meta:
        abstract = True

    def __str__(self):
        return self.full_address or self.get_address()

    def get_address(self):
        if self.address and self.city and self.state and self.country and self.postal_code:
            return "%s, %s, %s %s, %s" % (self.address, self.city, self.state, self.postal_code, self.country)

        return None



import django_filters


def get_filterset_base(Model):
    class MediaModelFilter(django_filters.FilterSet):
        modified__gt = django_filters.IsoDateTimeFilter(field_name='modified', lookup_expr='gt')

        class Meta:
            model = Model
            fields = {
                'id': ['exact'],
                "title": ['icontains'],
                "description": ['icontains'],
                "modified": ['exact', 'gt'],
                "created": ['exact'],
                "status": ['exact'],
                "parent_id": ['exact'],
                "parent_type_id": ['exact'],
                "category": ['exact'],
                "origin_url": ['exact'],
                "origin_id": ['exact'],
                "content_type": ['exact'],
            }

    return MediaModelFilter



from django import forms
from django.db.models import Field, CharField
from django.db.models.signals import pre_save
from django.utils.translation import gettext as _

from bitsoframework.utils.strings import random


class RandomIDField(CharField):
    default_error_messages = {
        'invalid': _("'%(value)s' is not a valid ID."),
    }
    description = 'Unique identifier'
    empty_strings_allowed = False

    def __init__(self, verbose_name=None, **kwargs):
        if not "max_length" in kwargs:
            kwargs["max_length"] = 7
        kwargs["primary_key"] = True
        kwargs["default"] = self.generate_id
        super().__init__(verbose_name, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        del kwargs['max_length']
        del kwargs['primary_key']
        del kwargs['default']
        return name, path, args, kwargs

    def validate_id(self, value):
        return not self.model.objects.filter(pk=value).exists()

    def generate_id(self):
        return random(self.max_length, self.validate_id)

    def get_db_prep_value(self, value, connection=None, prepared=False):
        if not value:
            value = self.generate_id()
        return value

    def validate(self, value, model_instance):
        pass

    def get_prep_value(self, value):
        value = super().get_prep_value(value)
        return value

    def formfield(self, **kwargs):
        return None

    def contribute_to_class(self, cls, name, **kwargs):
        assert not cls._meta.auto_field, "Model %s can't have more than one RandomIDField." % cls._meta.label
        super().contribute_to_class(cls, name, **kwargs)
        cls._meta.auto_field = self

        def on_before_save(sender, **kwargs):
            instance = kwargs.get("instance")
            if instance and not getattr(instance, name):
                new_id = self.generate_id()
                setattr(name, new_id)

        pre_save.connect(on_before_save, sender=cls)

    def to_python(self, value):
        return value



# Generated by Django 2.1.5 on 2020-02-18 13:25

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django_extensions.db.fields
import jsonfield.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Device',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', django_extensions.db.fields.CreationDateTimeField(auto_now_add=True, verbose_name='created')),
                ('modified', django_extensions.db.fields.ModificationDateTimeField(auto_now=True, verbose_name='modified')),
                ('status', models.IntegerField(choices=[(0, 'Inactive'), (1, 'Active')], default=1, verbose_name='status')),
                ('activate_date', models.DateTimeField(blank=True, help_text='keep empty for an immediate activation', null=True)),
                ('deactivate_date', models.DateTimeField(blank=True, help_text='keep empty for indefinite activation', null=True)),
                ('service', models.CharField(choices=[('firebase-push', 'firebase-push'), ('firebase-web-push', 'firebase-web-push')], db_index=True, max_length=300)),
                ('token', models.CharField(db_index=True, max_length=300)),
                ('model', models.CharField(blank=True, db_index=True, max_length=300, null=True)),
                ('platform', models.CharField(blank=True, db_index=True, max_length=300, null=True)),
                ('serial', models.CharField(blank=True, max_length=300, null=True)),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='devices', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'db_table': 'bitso_device',
            },
        ),
        migrations.CreateModel(
            name='DeviceNotification',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', django_extensions.db.fields.CreationDateTimeField(auto_now_add=True, verbose_name='created')),
                ('modified', django_extensions.db.fields.ModificationDateTimeField(auto_now=True, verbose_name='modified')),
                ('title', models.CharField(max_length=255, verbose_name='title')),
                ('description', models.TextField(blank=True, null=True, verbose_name='description')),
                ('status', models.IntegerField(choices=[(0, 'Inactive'), (1, 'Active')], default=1, verbose_name='status')),
                ('activate_date', models.DateTimeField(blank=True, help_text='keep empty for an immediate activation', null=True)),
                ('deactivate_date', models.DateTimeField(blank=True, help_text='keep empty for indefinite activation', null=True)),
                ('data', jsonfield.fields.JSONField(null=True)),
                ('device', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='notifications', to='notifications.Device')),
                ('recipient', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='device_notifications', to=settings.AUTH_USER_MODEL)),
                ('sender', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='sent_device_notifications', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'db_table': 'bitso_device_notification',
            },
        ),
    ]



from bitsoframework.media.services import MediaService
from bitsoframework.media.settings import CloudDocument


class CloudDocumentService(MediaService):
    """
    Service used to manage the lifecycle of the Cloud document model.
    """

    model_class = CloudDocument



from django.contrib import admin

from bitsoframework.media.settings import CloudDocument


class CloudDocumentAdmin(admin.ModelAdmin):
    model = CloudDocument
    list_display = ["id", "title", "origin_url", "origin_type"]
    search_fields = ["title", "origin_url", "origin_type"]



punctuations = ['(', ')', ';', ':', '[', ']', ',']
stop_words = ['', ' ']


def get_words(text):
    tokens = text.split(" ")
    keywords = [word for word in tokens if not word in stop_words and not word in punctuations]
    return keywords



from django.contrib.auth import logout
from rest_framework.decorators import action
from rest_framework.response import Response


class LogoutViewMixin(object):

    @action(methods=["post"], detail=False, permission_classes=[])
    def logout(self, request):
        logout(request)

        return Response("OK")


class UpdatePasswordViewMixin(object):
    @action(methods=["post"], detail=False)
    def update_password(self, request):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        serializer.update_password()

        return Response("OK")


class ForgotPasswordViewMixin(object):
    @action(methods=["post"], detail=False, permission_classes=[])
    def forgot_password(self, request):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        serializer.forgot_password()

        return Response("OK")


class ResetPasswordViewMixin(object):
    @action(methods=["post"], detail=False, permission_classes=[])
    def reset_password(self, request):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        serializer.reset_password()

        return Response("OK")



from django_extensions.db.models import ActivatorModel
from rest_framework.decorators import action
from rest_framework.response import Response
from rest_framework.status import HTTP_201_CREATED, HTTP_200_OK

from bitsoframework.notifications.serializers import DeviceSerializer


class DeviceViewMixin(object):

    @action(methods=["GET"], detail=False, url_path="devices")
    def list_devices(self, request):
        """
        Attach a new photo to the underlying record.
        """

        queryset = request.user.devices.active()

        serializer = DeviceSerializer(instance=queryset, many=True)

        return Response(serializer.data)

    @action(methods=["POST"], detail=False, url_path="devices/save")
    def register_device(self, request):
        """
        Attach a new photo to the underlying record.
        """

        instance = request.user.devices.filter(token=request.data.get("token")).first()

        if instance:

            serializer = DeviceSerializer(data=request.data, instance=instance, partial=True)
            serializer.is_valid(raise_exception=True)

            serializer.save(status=ActivatorModel.ACTIVE_STATUS)

            return Response(serializer.data, status=HTTP_200_OK)
        else:

            serializer = DeviceSerializer(data=request.data)
            serializer.is_valid(raise_exception=True)

            serializer.save(user=request.user)

            return Response(serializer.data, status=HTTP_201_CREATED)



from rest_framework.decorators import action
from rest_framework.response import Response

from bitsoframework.serializers import ListIDSerializer


class IndexedModelMixin(object):

    def get_serializer_class(self):
        if self.action == 'reorder':
            return ListIDSerializer

        return super(IndexedModelMixin, self).get_serializer_class()

    @action(detail=False, methods=["POST"])
    def reorder(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        self.get_queryset().model.objects.reorder(serializer.data.get("ids"))

        return Response("OK")



from django.db.models import QuerySet, Manager, Model, IntegerField
from django_extensions.db.models import ActivatorModelManager, ActivatorQuerySet


class IndexedQuerySet(QuerySet):

    def indexed(self):
        return self.order_by("index")


class IndexedModelManager(Manager):

    def get_queryset(self):
        return IndexedQuerySet(self.model, using=self._db)

    def reorder(self, keys):
        """
        Re-order all the items mapped the given indexed map of keys

        @param Class the class type we are re-ordering

        @param keys the map holding the pk -> index values
        """

        fields = ["index"]

        if isinstance(keys, (list, tuple)):

            index = 0

            for key in keys:
                model = self.model(id=key)
                model.index = index
                model.save(update_fields=fields)

                index += 1

        elif isinstance(keys, dict):

            for key in keys.keys():
                model = self.model(id=key)
                model.index = keys.get(key)
                model.save(update_fields=fields)

        else:
            raise Exception(
                "Expecting list/tuple of identifiers or dictionary of {id=index} but instead got: " + keys.__class__.name)


class IndexedModel(Model):
    index = IntegerField(null=True, blank=True, db_index=True)
    objects = IndexedModelManager()

    class Meta:
        abstract = True


# Utilities

class ActivatorIndexedQuerySet(ActivatorQuerySet, IndexedQuerySet):
    pass


class ActivatorIndexedModelManager(ActivatorModelManager, IndexedModelManager):

    def get_queryset(self):
        return ActivatorIndexedQuerySet(self.model, using=self._db)



from bitsoframework.utils.models import get_model_class, get_field, exists, diff_list


class Differ(object):

    def __init__(self, record):
        self.model_class = get_model_class(record)
        self.new_record = record
        self.old_record = self.model_class.objects.get(pk=record.pk) if exists(record) else None

    def compare(self, properties):

        result = {}

        for name in properties:
            field = get_field(self.model_class, name) or getattr(self.model_class, name)
            new_value = getattr(self.new_record, name)
            old_value = getattr(self.old_record or self.new_record, name)

            if new_value and new_value.__class__.name == "RelatedManager":

                old_value = list(getattr(self.old_record, name).all()) if self.old_record else []
                new_value = list(new_value.all())

                result[name] = {
                    "old_value": old_value,
                    "new_value": new_value,
                    **diff_list(old_value, new_value)
                }

            elif new_value != old_value:

                result[name] = {
                    "old_value": old_value,
                    "new_value": new_value
                }

        return result



def to_string(cls, exception):
    """
    Translate the given exception to String, either by its defined message or
    """

    if isinstance(exception, Exception):

        if hasattr(exception, 'message') and exception.message:

            return exception.message

        else:

            return str(exception)

    if exception is not None:
        return str(exception)

    return None



from .settings import ON_DELETE, AUTH_USER_MODEL
from django.db import models
from django.db.models import Model


class AuthoredModel(Model):
    """
    Model that tracks users who have craeted or last updated a record.

    Note that by itself, this model only holds the two properties but it does not
    automatically set them. You must use AuthoredSerializer in conjunction with
    your REST API or manually specify them instead.

    This model replaced CreatorModel in favor of decoupled logic.
    """

    class Meta:
        abstract = True

    created_by = models.ForeignKey(AUTH_USER_MODEL, on_delete=ON_DELETE,
                                   related_name="%(app_label)s_%(class)s_created", null=True,
                                   blank=True)
    """
    Reference to the user who created this record.
    """

    modified_by = models.ForeignKey(AUTH_USER_MODEL, on_delete=ON_DELETE,
                                    related_name="%(app_label)s_%(class)s_modified",
                                    null=True, blank=True)
    """
    Reference to the user who last updated this record.
    """



import tempfile

import googlemaps
from django.conf import settings
from django.utils.functional import cached_property


class GoogleMapService(object):

    def __init__(self, client_id=None, api_key=None):

        super(GoogleMapService, self).__init__()

        self._client = None
        self.client_id = client_id or getattr(settings, "GOOGLE_MAPS_CLIENT_ID", None)
        self.client_api_key = api_key or getattr(settings, "GOOGLE_MAPS_API_KEY", None)

    @cached_property
    def client(self):
        return googlemaps.Client(self.client_api_key,
                                 client_id=self.client_id)

    def update_coordinates(self, record):

        coordinates = self.get_coordinates(record)

        if record.coordinates != coordinates:
            record.coordinates = coordinates
            return True

        return False

    def get_address_query(self, record):

        if record:
            return record.get_address()

        return None

    def get_place_photo(self, *args, **kwargs):

        local_filename = tempfile.mktemp()

        f = open(local_filename, 'wb')

        for chunk in self.client.places_photo(*args, **kwargs):
            if chunk:
                f.write(chunk)
        f.close()

        return local_filename

    def get_directions(self, origin, destination):

        directions = self.client.directions(mode="driving", origin={
            "latitude": origin.y,
            "longitude": origin.x
        }, destination={
            "latitude": destination.y,
            "longitude": destination.x
        })

        return directions

    def get_coordinates(self, address):

        query = self.get_address_query(address)

        if query:
            geocode = self.client.geocode(query)
            if geocode and len(geocode) > 0:
                location = geocode[0]["geometry"]["location"]
                from django.contrib.gis.geos.point import Point
                return Point(x=location.get("lng"),
                             y=location.get("lat"))

        return None



from rest_framework.routers import DefaultRouter

from bitsoframework.metadata.views import *

router = DefaultRouter()

# model
router.register(r'model', MetadataModelViewSet, basename='metadata-model')

urlpatterns = router.urls



from rest_framework.decorators import action
from rest_framework.response import Response

from bitsoframework.metadata.serializers import MetadataModelSerializer
from bitsoframework.metadata.services import MetadataModelService
from bitsoframework.mixins import AbstractViewSet


class MetadataModelViewSet(AbstractViewSet):
    serializer_class = MetadataModelSerializer

    def __init__(self, **kwargs):
        super(MetadataModelViewSet, self).__init__(**kwargs)
        self.service = MetadataModelService()

    def get(self):
        pass

    @action(detail=False, url_path="(?P<app_label>[-\w]+)/(?P<model_name>[-\w]+)")
    def get_by_name(self, request, app_label, model_name):
        instance = self.service.get_model(app_label, model_name)

        serializer = self.get_serializer(instance=instance)

        return Response(serializer.data)



from dateutil.parser import parse

class Dates(object):

    @classmethod
    def is_date(string, fuzzy=False):
        """
        Return whether the string can be interpreted as a date.

        :param string: str, string to check for date
        :param fuzzy: bool, ignore unknown tokens in string if True
        """
        try:
            parse(string, fuzzy=fuzzy)
            return True

        except ValueError:
            return False


"""
Media path resolver class that takes a django template to configure where files should be uploaded
within the media repository.
"""

from django.template.base import Template
from django.template.context import Context
from django.utils.deconstruct import deconstructible

from bitsoframework.media.models import MEDIA_CATEGORY_CHOICES
from bitsoframework.utils.models import get_choice_label


@deconstructible()
class MediaPathResolver(object):
    parser = None

    def __init__(self, path=None,
                 template="""
{% if path %}
    {{path}}/
{% endif %}
{% if parent_type %}
    {{parent_type}}/
{% endif %}
{% if parent_id %}
    {{parent_id}}/
{% endif %}
{% if category %}
    {{category}}/
{% endif %}
{{filename}}
"""):
        self.path = path
        self.template = template

    def __call__(self, instance, filename, allow_override=True):

        response = instance.resolve_path(filename, self)

        if not response:

            if not self.parser:
                self.parser = Template(self.template)

            category = None

            if instance.category:
                category = get_choice_label(MEDIA_CATEGORY_CHOICES, instance.category)

            if instance.parent_type:
                parent_type = instance.parent_type.model

                parent_id = instance.parent_id

            context = locals()
            context["path"] = self.path

            response = self.parser.render(Context(context))
            response = response.replace("\n", "").replace("\t", "\n").replace("    ", "").strip()

        return response



import hashlib
from django.core.files.base import File


def calculate_md5(media):
    return calculate(media, "md5")


def calculate_sha1(media):
    return calculate(media, "sha1")


def calculate_sha224(media):
    return calculate(media, "sha224")


def calculate_sha256(media):
    return calculate(media, "sha256")


def calculate_sha384(media):
    return calculate(media, "sha384")


def calculate_sha512(media):
    return calculate(media, "sha512")


def calculate(media, type):
    file = get_file(media)

    if file:
        hash = getattr(hashlib, type)()

        content = file.read()

        hash.update(content)

        digest = hash.hexdigest()

        return digest

    return None


def get_file(media):
    if isinstance(media, File):
        return media

    if isinstance(media, str):
        return open(media, "rb")

    if hasattr(media, "get_file"):
        return media.get_file()

    if hasattr(media, "file"):
        return media.file

    return None



import datetime

from django_extensions.db.models import ActivatorModel, TimeStampedModel
from rest_framework.decorators import action
from rest_framework.response import Response

from bitsoframework.serializers import ListIDSerializer
from bitsoframework.utils.models import set_attrs


class ActivateModelMixin(object):

    def get_serializer_class(self):
        if self.action == 'activate':
            return ListIDSerializer

        return super(ActivateModelMixin, self).get_serializer_class()

    @action(detail=False, methods=["POST"])
    def activate(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        records = self.get_queryset().filter(id__in=serializer.data.get("ids"))

        data = {
            "status": ActivatorModel.ACTIVE_STATUS,
            "activate_date": datetime.datetime.now()
        }

        if issubclass(records.model, TimeStampedModel):
            data["modified"] = datetime.datetime.now()

        # TODO: find a way to trigger signals on queryset natively...

        for record in records:
            set_attrs(record, data)
            record.save(update_fields=data.keys())

        # records.update(**data)

        # for record in records:
        #    post_save.send(record.__class__, instance=record, created=False)

        return Response("OK")


class DeactivateModelMixin(object):

    def get_serializer_class(self):
        if self.action == 'deactivate':
            return ListIDSerializer

        return super(DeactivateModelMixin, self).get_serializer_class()

    @action(detail=False, methods=["POST"])
    def deactivate(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        records = self.get_queryset().filter(id__in=serializer.data.get("ids"))

        data = {
            "status": ActivatorModel.INACTIVE_STATUS,
            "deactivate_date": datetime.datetime.now()
        }

        if issubclass(records.model, TimeStampedModel):
            data["modified"] = datetime.datetime.now()

        # TODO: find a way to trigger signals on queryset natively...
        for record in records:
            set_attrs(record, data)
            record.save(update_fields=data.keys())

        # records.update(**data)
        # for record in records:
        #    post_save.send(records.model, instance=record, created=False)

        return Response("OK")


class ActiveModelMixin(object):

    @action(detail=False)
    def active(self, request, *args, **kwargs):
        """
        List active queryset.
        """
        queryset = self.filter_queryset(self.get_queryset().filter(status=ActivatorModel.ACTIVE_STATUS))

        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = self.get_serializer(page, many=True)
            return self.get_paginated_response(serializer.data)

        serializer = self.get_serializer(queryset, many=True)
        return Response(serializer.data)


class InactiveModelMixin(object):

    @action(detail=False)
    def inactive(self, request, *args, **kwargs):
        """
        List inactive queryset.
        """
        queryset = self.filter_queryset(self.get_queryset().filter(status=ActivatorModel.INACTIVE_STATUS))

        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = self.get_serializer(page, many=True)
            return self.get_paginated_response(serializer.data)

        serializer = self.get_serializer(queryset, many=True)
        return Response(serializer.data)


class ActivatorModelMixin(ActivateModelMixin,
                          DeactivateModelMixin,
                          ActiveModelMixin,
                          InactiveModelMixin):
    pass



from bitsoframework.metadata.factory.base import AbstractModelFactory


class MetadataModelSerializerFactory(AbstractModelFactory):

    serializers = None

    def get_model(self, app_label, model_name):
        pass



from django.conf import settings as django_settings
from django.db.models import fields, ForeignKey, CASCADE, SET_NULL
from django.utils.translation import gettext as _
from django_extensions.db.models import TitleDescriptionModel, TimeStampedModel, ActivatorModel
from jsonfield import JSONField

from bitsoframework.notifications import settings


class Device(TimeStampedModel, ActivatorModel):
    DEVICE_SERVICE_CHOICES = [(key, _(key)) for key in settings.NOTIFICATION_SERVICES.keys()]

    user = ForeignKey(django_settings.AUTH_USER_MODEL, on_delete=CASCADE, related_name="devices")
    service = fields.CharField(max_length=300, db_index=True, choices=DEVICE_SERVICE_CHOICES)
    manufacturer = fields.CharField(max_length=300, null=True, blank=True)
    model = fields.CharField(max_length=300, db_index=True, null=True, blank=True)
    platform = fields.CharField(max_length=300, db_index=True, null=True, blank=True)
    serial = fields.CharField(max_length=300, null=True, blank=True)
    token = fields.CharField(max_length=300, db_index=True, null=True, blank=True)
    uuid = fields.CharField(max_length=300, null=True, blank=True)
    version = fields.CharField(max_length=300, null=True, blank=True)

    class Meta:
        db_table = "bitso_device"

    def __str__(self):
        return "Device(id=%s, user=%s, service=%s, manufacturer=%s, model=%s, platform=%s, serial=%s, token=%s, uuid=%s, version=%s)" % \
               (str(self.id), str(self.user), str(self.service), str(self.manufacturer), str(self.model),
                str(self.platform), str(self.serial), str(self.token), str(self.uuid), str(self.version))


class DeviceNotification(TitleDescriptionModel, TimeStampedModel, ActivatorModel):
    recipient = ForeignKey(django_settings.AUTH_USER_MODEL, on_delete=CASCADE, null=True,
                           related_name="device_notifications")
    sender = ForeignKey(django_settings.AUTH_USER_MODEL, on_delete=CASCADE, null=True,
                        related_name="sent_device_notifications")
    device = ForeignKey(Device, on_delete=SET_NULL, null=True, related_name="notifications")
    data = JSONField(null=True)

    class Meta:
        db_table = "bitso_device_notification"

    def __str__(self):
        return "DeviceNotification(id=%s, title=%s, description=%s, data=%s, recipient=%s, device=%s)" % \
               (str(self.id), str(self.title), str(self.description), str(self.data), str(self.recipient),
                str(self.device))



import tempfile
import os
from uuid import uuid4

import mimetypes

"""
File utilities and tricks

@author: bitsoframework

@since: 07/25/2014
"""


def save(path, content):
    f = open(path, "w")
    f.write(content)
    f.close()


def list(directory, include_files=True, include_directories=False, recursive=False):
    """
    Build up a list with files and directories under the given directory or
    according to the given filter parameterized in this method call.

    @return: list of files
    """
    files = list()

    for filename in os.listdir(directory):

        file_path = os.path.join(directory, filename)

        if os.path.isfile(file_path):

            if include_files:
                files.append(file_path)

        elif os.path.isdir(file_path):

            if include_directories:
                files.append(file_path)

            if recursive:

                for f in list(filename, include_files, include_directories, recursive):
                    files.add(f)

    return files


def delete_file(file_path):
    return os.remove(file_path)


def read_base64(file_path):
    return read(file_path).encode("base64")


def read(file_path):
    f = open(file_path, 'rb')
    s = f.read()
    f.close()
    return s


def get_remote_content_type(url):
    from urllib.request import urlopen
    try:
        res = urlopen(url)
        http_message = res.info()
        return http_message.get_content_type()
    except:
        if url and url.startswith("http"):
            return "text/html"

        return None


def get_content_type(filename):
    return mimetypes.MimeTypes().guess_type(filename)[0]


def download(url, output=None, debug=False):
    """
    Download and save a file specified by url to output file. If not provided,
    creates a temp file instead.
    """
    from urllib.request import urlopen

    u = urlopen(url)

    if not output:
        output = tempfile.mktemp()

    with open(output, 'wb') as f:
        meta = u.info()
        meta_func = meta.getheaders if hasattr(meta, 'getheaders') else meta.get_all
        meta_length = meta_func("Content-Length")
        file_size = None
        if debug and meta_length:
            file_size = int(meta_length[0])
            print("Downloading: {0} Bytes: {1}".format(url, file_size))

        file_size_dl = 0
        block_sz = 8192
        while True:
            buffer = u.read(block_sz)
            if not buffer:
                break

            file_size_dl += len(buffer)
            f.write(buffer)

            if debug:
                status = "{0:16}".format(file_size_dl)
                if file_size:
                    status += "   [{0:6.2f}%]".format(file_size_dl * 100 / file_size)
                status += chr(13)

                print(status),
                print()

    return output


def get_extension(file_path):
    """
    Utility used to extract the extension of a given file path.

    :param file_path:
    :return:
    """
    name, extension = os.path.splitext(file_path)

    return extension[1:].lower()


def get_filename(filename):
    """
    Utility used to extract the extension of a given file path.

    :param filename:
    :return:
    """
    name, extension = os.path.splitext(filename)

    return name


def get_random_name(name):
    """
    Generate a random name for the given file name preserving the original file's
    name extension

    @param name the file name

    @return a random file name
    """

    ext = os.path.splitext(name)[1]

    return uuid4().__str__() + ext



from django.conf import settings
from django.db import models
from django.db.models import SET_NULL
from django.db.models.fields.files import ImageFieldFile
from django.utils.functional import cached_property
from easy_thumbnails.files import ThumbnailFile

from bitsoframework.media.models import MEDIA_PATH_RESOLVER, AbstractFileMedia
from bitsoframework.utils.reflection import exists

THUMBNAIL_ENABLED = getattr(settings, "BITSO_IMAGE_THUMBNAIL_ENABLED", exists("easy_thumbnails.models.Source"))
THUMBNAIL_SOURCE_MODEL = getattr(settings, "BITSO_IMAGE_THUMBNAIL_SOURCE", "easy_thumbnails.Source")


class AbstractImage(AbstractFileMedia):
    """
    Model used to map an image uploaded/attached to another model in the system.

    @since: 06/17/2014 20:30:00

    @author: bitsoframework
    """

    type = "image"
    """
    The unique type of document among the various registered document types.
    """

    file = models.ImageField('The image itself', upload_to=MEDIA_PATH_RESOLVER, width_field="width",
                             height_field="height", max_length=500)
    """
    The image itself
    """

    if THUMBNAIL_ENABLED:
        thumbnails_source = models.OneToOneField(THUMBNAIL_SOURCE_MODEL, null=True, blank=True, on_delete=SET_NULL,
                                                 related_name="image")
        """
        The easy_thumbnail source for this file so we can encapsulate all of its thumbnails into a single db reference.
        """

    thumbnails_filesize = models.IntegerField(null=True, blank=True)
    """
    The number of bytes for all the generated thumbnails.
    """

    width = models.FloatField(null=True, blank=True)
    """
    The width (in pixels) for the image
    """

    height = models.FloatField(null=True, blank=True)
    """
    The height (in pixels) for the image
    """

    class Meta:
        abstract = True

    @cached_property
    def thumbnail_aliases(self):

        data = getattr(settings, "THUMBNAIL_ALIASES", {})

        aliases = []

        for value in data.values():

            for name in value.keys():
                aliases.append(name)

        return aliases

    @cached_property
    def thumbnails(self):
        from easy_thumbnails.files import get_thumbnailer
        from easy_thumbnails.alias import aliases

        data = {}
        thumbnails = list(self.thumbnails_source.thumbnails.all()) if self.thumbnails_source else []
        thumbnailer = get_thumbnailer(self.file)

        if len(thumbnails) != len(aliases.all()):
            self.generate_thumbnails()
            thumbnails = list(self.thumbnails_source.thumbnails.all())

        for thumbnail in thumbnails:
            for alias in aliases.all():

                options = aliases.get(alias, target=thumbnailer.alias_target)
                thumbnail_options = thumbnailer.get_options(options)

                names = [
                    thumbnailer.get_thumbnail_name(
                        thumbnail_options, transparent=False,
                        high_resolution=False)]
                transparent_name = thumbnailer.get_thumbnail_name(
                    thumbnail_options, transparent=True,
                    high_resolution=False)
                if transparent_name not in names:
                    names.append(transparent_name)

                for name in names:
                    if thumbnail.name == name:
                        data[alias] = ThumbnailFile(
                            name=name, storage=thumbnailer.thumbnail_storage,
                            thumbnail_options=thumbnail_options)
                        break

        return data

    def thumbnail_by_alias(self, alias):
        return self.thumbnails.get(alias)

    @property
    def urls(self):

        urls = {
            "original": self.url
        }

        if THUMBNAIL_ENABLED:
            for alias, thumbnail in self.thumbnails.items():
                urls[alias] = thumbnail.url

        return urls

    def on_update_file(self):

        if THUMBNAIL_ENABLED:
            self.thumbnails_source = None

        super(AbstractImage, self).on_update_file()

    def delete_file(self, _file):

        self.thumbnails_filesize = None

        if isinstance(_file, ImageFieldFile) and _file.name:

            self.storage.delete(_file.name)

        elif isinstance(_file, str):

            self.storage.delete(_file)

        try:
            from easy_thumbnails.files import get_thumbnailer

            thumbnailer = get_thumbnailer(_file)
            thumbnailer.delete_thumbnails()

        except ImportError:
            pass

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):

        creating = not self.id

        super(AbstractImage, self).save(force_insert=force_insert, force_update=force_update, using=using,
                                        update_fields=update_fields)

        if creating and THUMBNAIL_ENABLED:
            self.generate_thumbnails()

    def generate_thumbnails(self):

        from easy_thumbnails.templatetags import thumbnail
        from easy_thumbnails.models import Source
        aliases = getattr(settings, "THUMBNAIL_ALIASES", {})

        for value in aliases.values():

            for name in value.keys():
                thumbnail.thumbnail_url(self.file, name)

        if not self.thumbnails_source:
            self.thumbnails_source = Source.objects.get_file(self.storage, self.file.name)
            self.save(update_fields=['thumbnails_source'])

    def copy(self, **kwargs):

        if THUMBNAIL_ENABLED:
            kwargs["thumbnails_source"] = None

        copy = super(AbstractImage, self).copy(**kwargs)

        return copy

    def resolve_path(self, filename, resolver):

        if hasattr(self, "preview_source") and self.preview_source:
            return resolver(self.preview_source, "preview/" + filename, allow_override=False)

        return super(AbstractImage, self).resolve_path(filename, resolver)



from rest_framework.fields import CharField
from rest_framework.serializers import Serializer


class MetadataAttributeSerializer(Serializer):
    name = CharField()
    type = CharField()


class MetadataModelSerializer(Serializer):
    name = CharField()
    attributes = MetadataAttributeSerializer(many=True)



import base64
import binascii
import imghdr
import mimetypes
import uuid

from PIL import Image as PILImage
from django.core.exceptions import ValidationError
from django.core.files.base import ContentFile, File
from django.utils.encoding import force_bytes
from django.utils.http import urlsafe_base64_encode
from django.utils.translation import ugettext_lazy as _

from bitsoframework.middleware import get_request
from bitsoframework.utils import files

EMPTY_VALUES = (None, '', [], (), {})
DEFAULT_CONTENT_TYPE = "application/octet-stream"
ALLOWED_IMAGE_TYPES = (
    "jpe",
    "jpeg",
    "jpg",
    "png",
    "gif"
)


def download(url):
    file_path = files.download(url)

    return File(file=open(file_path, 'rb'), name=file_path)


def get_uid(media):
    return "%s:%s" % (media.type, urlsafe_base64_encode(force_bytes(media.id)))


def get_files(request=None):
    request = request or get_request()

    files = []

    for key in request.FILES.keys():

        for file in request.FILES.getlist(key):
            files.append(file)

    return files


def is_image(file):
    try:
        image = PILImage.open(file)
        image.verify()
        # image.close()
        return True
    except:
        return False


class Base64:
    """
    Utility used to parse a base64 into a temporary file
    """

    @classmethod
    def get_file(cls, base64_data):
        """
        convert the given base64 data to a file and file extension

        :param base64_data:
        :return:
        """

        # Check if this is a base64 string
        if base64_data in EMPTY_VALUES:
            return None, None

        if isinstance(base64_data, str):
            # Strip base64 header.
            if ';base64,' in base64_data:
                header, base64_data = base64_data.split(';base64,')

            # Try to decode the file. Return validation error if it fails.
            try:
                decoded_file = base64.b64decode(base64_data)
            except (TypeError, binascii.Error):
                raise ValidationError(_("Please upload a valid image."))

            # Generate file name:
            file_name = str(uuid.uuid4())[:12]  # 12 characters are more than enough.

            # Get the file name extension:
            file_extension = Base64.get_file_extension(file_name, decoded_file, header.replace("data:", ""))

            if file_extension not in ALLOWED_IMAGE_TYPES:
                raise ValidationError(_("The type of the image couldn't been determined."))

            complete_file_name = file_name + "." + file_extension

            data = ContentFile(decoded_file, name=complete_file_name)

            return data, file_extension

        raise ValidationError(_('This is not an base64 string'))

    @classmethod
    def get_file_extension(cls, filename, decoded_file, content_type):

        extension = imghdr.what(filename, decoded_file)
        extension = "jpg" if extension == "jpeg" else extension

        if extension is None and content_type:
            extension = mimetypes.guess_extension(content_type)
            if extension:
                extension = extension.replace(".", "")

        return extension



from django.conf import settings
from django.db import models
from django.db.models import SET_NULL
from django.db.models.fields.files import FieldFile

from bitsoframework.media.models import MEDIA_PATH_RESOLVER, AbstractFileMedia

IMAGE_MODEL_NAME = getattr(settings, "BITSO_MEDIA_DOCUMENT_MODEL_NAME", "media.Image")


class AbstractDocument(AbstractFileMedia):
    """
    Model used to map an image uploaded/attached to another model in the system.

    @since: 06/17/2016 20:30:00

    @author: bitsoframework
    """

    type = "document"
    """
    The unique type of document among the various registered document types.
    """

    file = models.FileField('The document itself', upload_to=MEDIA_PATH_RESOLVER, max_length=500)
    """
    The image itself
    """

    preview = models.OneToOneField(IMAGE_MODEL_NAME, on_delete=SET_NULL, null=True, blank=True,
                                   related_name="preview_source")
    """
    If preview tools are installed, this stores the reference of the image used to preview the document's content. 
    """

    class Meta:
        abstract = True

    def delete_file(self, _file):

        if isinstance(_file, FieldFile):

            self.storage.delete(_file.name)

        elif isinstance(_file, str):

            self.storage.delete(_file)



#
# Example:
#
# on routers.py
#
# from bitsoframework.db.routers import DatabaseRouter
#
# class YourRouter(DatabaseRouter):
#
#    db_name = "your_db_name"
#    models = [...]
#    modules = [...]
#
# on settings.py add
#
# DATABASE_ROUTERS = ['xxx.routers.YourRouter']
#

class DatabaseRouter(object):
    """
    A router to control database routing so specific models or modules are
    directed to a specific table

    @author: bitsoframework
    """

    db_name = None
    models = []
    modules = []

    def match(self, model):

        if self.modules.__contains__(model._meta.app_label):
            return True

        if self.models.__contains__(model._meta.model_name):
            return True

        return False

    def db_for_read(self, model, **hints):

        if self.match(model):
            return self.db_name

        return None

    def db_for_write(self, model, **hints):

        if self.match(model):
            return self.db_name

        return None

    def allow_relation(self, obj1, obj2, **hints):

        if self.match(obj1) and self.match(obj2):
            return True

        return None

    def allow_migrate(self, db, app_label, model=None, **hints):
        if db == self.db_name:
            return False

        return None



class BatchManager(object):

    def __init__(self):
        self.actions = []

    def register(self, action):
        self.actions.append(action)

    def __call__(self, *args, **kwargs):
        self.run()

    def run(self):
        for action in self.actions:
            action()

        self.actions = []



import PyPDF2
import textract
from django.core.files import File


def get_text(file_path, language="eng", preferences=["textract", "pypdf"], include_all=False):
    """
    Extract text for all pages of the given PDF document
    """
    if isinstance(file_path, File):
        f = file_path.open('rb')
        file_path = f.name
    elif isinstance(file_path, str, ):
        f = open(file_path, 'rb')
    else:
        f = file_path

    text = None

    result = []

    for preference in preferences:

        if preference == "textract":

            text = textract.process(file_path, method='tesseract', language=language).decode("UTF-8")

        elif preference == "pypdf":

            pdf = PyPDF2.PdfFileReader(f)

            num_pages = pdf.numPages
            count = 0
            text = ""

            while count < num_pages:
                page = pdf.getPage(count)
                count += 1
                text += page.extractText()

        if text:
            if include_all:
                result.append(text)

            else:
                return text

    return result if include_all else None



def require(name):
    if isinstance(name, str):
        index = name.rfind(".")
        module_name = name[0:index]
        member_name = name[index + 1:len(name)]
        module = __import__(module_name, globals(), locals(), [member_name])

        return getattr(module, member_name)

    return name


def exists(name):
    try:
        return require(name) is not None
    except:
        return False



import os

import firebase_admin
from django.conf import settings
from firebase_admin import credentials

cred_file = getattr(settings, 'BITSO_FIREBASE_CREDENTIALS',
                    os.path.join(settings.BASE_DIR, "keys/firebase.json"))
cred = credentials.Certificate(cred_file)
app = firebase_admin.initialize_app(cred)



from bitsoframework.metadata.models import MetadataModel


class AbstractModelFactory(object):

    def get_model(self, app_label, model_name):
        model = MetadataModel(app_label=app_label,
                              model_name=model_name)

        model.attributes = self.__build_attributes(model)

        return model

    def __build_attributes(self, model):
        return []



def as_list(value):
    if not isinstance(value, list):
        return list(value)

    return value


def add_all(array, other):
    """
    Append all items of the second array into the first one.

    :param array:
    :param other:
    :return:
    """

    for item in other:
        array.append(item)

    return array


def remove_all(array, other):
    """
    Remove all items of the second array into the first one.

    :param array:
    :param other:
    :return:
    """
    for item in other:
        array.remove(item)

    return array


def contains(array, *args):
    """
    Check rather the given array contains at least one of the given items
    :param array:
    :param args:
    :return: true if at least one item exists, false otherwise
    """

    for arg in args:

        if array.__contains__(arg):
            return True

    return False


def to_list(entries):
    if isinstance(entries, list):
        return entries

    values = list()

    if isinstance(entries, tuple):

        for entry in entries:
            values.append(entry)
    else:
        values.append(entries)

    return values


def flatten(*args):
    result = []

    for items in args:
        for item in items:
            result.append(item)

    return result



from bitsoframework.utils.models import set_attrs, get_choice_label
from bitsoframework.vos import AbstractVO


class Media(AbstractVO):
    """
    Value Object that represents a set of uploaded documents and images to a
    parent record.
    """

    categories = None
    """
    Required grouping categories.
    """

    def load(self, entries):
        """
        Load from a single collection of images and documents into this vo.
        """

        map = {}

        for entry in entries:

            name = get_choice_label(self.categories, entry.category)

            values = map.get(name, None)

            if not values:
                map[name] = values = list()

            values.append(entry)

        # index them
        for key, value in map.items():
            map[key] = sorted(value, key=lambda entry: entry.index, reverse=True)

        set_attrs(self, map)

        return self



from bitsoframework.media.image.models import THUMBNAIL_ENABLED
from bitsoframework.media.services import MediaService
from bitsoframework.media.settings import Image


class ImageService(MediaService):
    """
    Service used to manage the lifecycle of the Image model.
    """

    model_class = Image

    def filter(self, **kwargs):
        queryset = super(ImageService, self).filter(**kwargs)

        if THUMBNAIL_ENABLED:
            queryset = queryset.select_related("thumbnails_source").prefetch_related("thumbnails_source__thumbnails")

        return queryset



from bitsoframework.media.image.serializers import ImageSerializer
from bitsoframework.media.serializers import AbstractFileMediaSerializer
from bitsoframework.media.settings import Document, PREVIEW_ENABLED


class DocumentSerializer(AbstractFileMediaSerializer):
    if PREVIEW_ENABLED:
        preview = ImageSerializer(read_only=True)

    class Meta:
        model = Document
        exclude = ["parent_type", "parent_id", "file"]



from rest_framework import fields

from bitsoframework.media.serializers import AbstractFileMediaSerializer
from bitsoframework.media.settings import Image


class ImageSerializer(AbstractFileMediaSerializer):

    class Meta:
        model = Image
        exclude = ["parent_type", "parent_id", "file"]



from rest_framework import serializers
from rest_framework.serializers import ModelSerializer

from bitsoframework.notifications.models import Device


class DeviceSerializer(ModelSerializer):
    class Meta:
        model = Device
        exclude = ('user',)



from django.db import IntegrityError
from django.utils import timezone
from django_extensions.db.models import TimeStampedModel
from rest_framework import serializers, fields
from rest_framework.exceptions import ValidationError


class AnyField(fields.Field):

    def to_internal_value(self, data):
        return data

    def to_representation(self, value):
        return value


class ListIDSerializer(serializers.Serializer):
    ids = fields.ListField(child=AnyField(required=True))


class UpdateBulkListSerializer(serializers.ListSerializer):

    def update(self, instances, validated_data):

        instance_hash = {index: instance for index, instance in enumerate(instances)}

        result = [
            self.child.update(instance_hash[index], attrs)
            for index, attrs in enumerate(validated_data)
        ]

        writable_fields = [field.field_name for field in self.child._writable_fields]

        if isinstance(instances[0], TimeStampedModel):
            writable_fields.append("modified")
            modified = timezone.now()
            for instance in result:
                instance.modified = modified

        try:
            self.child.Meta.model.objects.bulk_update(result, writable_fields)
        except IntegrityError as e:
            raise ValidationError(e)

        return result



# Generated by Django 2.1.5 on 2020-02-21 13:45

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('notifications', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='device',
            name='manufacturer',
            field=models.CharField(blank=True, max_length=300, null=True),
        ),
        migrations.AddField(
            model_name='device',
            name='uuid',
            field=models.CharField(blank=True, max_length=300, null=True),
        ),
        migrations.AddField(
            model_name='device',
            name='version',
            field=models.CharField(blank=True, max_length=300, null=True),
        ),
        migrations.AlterField(
            model_name='device',
            name='token',
            field=models.CharField(blank=True, db_index=True, max_length=300, null=True),
        ),
    ]



from django.conf import settings
from django.utils.functional import cached_property
from django_extensions.db.models import TimeStampedModel, ActivatorModel
from firebase_admin import firestore

from bitsoframework.middleware import get_user, _active
from bitsoframework.utils import lists
from bitsoframework.utils.booleans import to_boolean


def firestore_batch_enabled():
    def decorator(func=None):
        def wrapper(*args, **kwargs):
            middleware = FirestoreBatchMiddleware(func)

            return middleware(*args, **kwargs)

        return wrapper

    return decorator


class FirestoreBatchMiddleware(object):

    def __init__(self, get_response):
        FirestoreChangeSynchronizer.batch_enabled = True

        self.get_response = get_response

    def __call__(self, *args, **kwargs):
        response = self.get_response(*args, **kwargs)

        FirestoreBatchMiddleware.batch_manager().commit()

        return response

    @classmethod
    def batch_manager(cls):
        if not hasattr(_active, "firestore_batch_manager"):
            _active.firestore_batch_manager = FirestoreBatchManager()

        return _active.firestore_batch_manager


class FirestoreBatch(object):
    _batch = None

    @cached_property
    def store(self):
        return firestore.client()

    @property
    def batch(self):
        if not self._batch:
            self._batch = self.store.batch()
        return self._batch

    def __init__(self):
        self.sets = {}

    def set(self, path, data, merge=False):

        self.sets[path] = {
            "data": data,
            "merge": merge
        }

    def commit(self):

        total = len(self.sets.keys())

        print("Committing %i transactions in firestore" % (total,))

        for path, item in self.sets.items():
            self.batch.set(self.store.document(path), item.get("data"), merge=item.get("merge"))

        self.batch.commit()


class FirestoreBatchManager(object):
    _batch = None

    @property
    def batch(self):
        if not self._batch:
            self._batch = FirestoreBatch()
        return self._batch

    def commit(self):
        if self._batch:
            self._batch.commit()
            self._batch = None


def log_change(collection, *args, **kwargs):
    handler = FirestoreChangeSynchronizer(collection=collection)
    handler.log(*args, **kwargs)
    return handler


class FirestoreChangeSynchronizer(object):
    enabled = getattr(settings, 'BITSO_FIRESTORE_SYNC_ENABLED', True)
    debug = getattr(settings, 'BITSO_FIRESTORE_SYNC_DEBUG', True)
    batch_enabled = getattr(settings, 'BITSO_FIRESTORE_SYNC_BATCH', False)
    snapshot_enabled = getattr(settings, 'BITSO_FIRESTORE_SYNC_SNAPSHOT', True)
    audit_enabled = getattr(settings, 'BITSO_FIRESTORE_SYNC_AUDIT', False)

    def __init__(self, collection, enabled=None, debug=None, batch_enabled=None, batch_manager=None):
        self.collection = collection
        self.transactions = []

        self.enabled = to_boolean(enabled, FirestoreChangeSynchronizer.enabled)
        self.debug = to_boolean(debug, FirestoreChangeSynchronizer.debug)
        self.batch_enabled = to_boolean(batch_enabled, FirestoreChangeSynchronizer.batch_enabled)
        self.batch_manager = batch_manager or (FirestoreBatchMiddleware.batch_manager() if self.batch_enabled else None)

    @cached_property
    def store(self):
        return firestore.client()

    def get_path(self, record=None, record_id=None):
        values = [str(value) for value in (lists.as_list(self.collection) + [record_id or record.id])]
        path = "/".join(values)
        if self.debug:
            print("Firestore path: " + path)
        return path

    def get(self, record=None, record_id=None):
        path = self.get_path(record, record_id)
        return self.store.document(path)

    def log(self, record, record_id=None, modified_by=None, is_delete=False, **kwargs):
        if not self.enabled:
            return

        record_id = record_id or record.id

        if not modified_by:
            user = get_user()
            if user:
                modified_by = user.id

        data = {
            "id": record_id,
            "modified_by": modified_by
        }

        if isinstance(record, ActivatorModel):
            data.update({
                'status': record.status
            })

        if isinstance(record, TimeStampedModel):
            data.update({
                'modified': record.modified
            })

        data.update(kwargs)

        if is_delete:
            data["deleted"] = True

        if self.batch_enabled:

            if self.audit_enabled:
                self.batch_manager.batch.set(self.get_path(record, record_id), data, merge=True)

            if self.snapshot_enabled:
                self.batch_manager.batch.set(self.get_path(record_id="snapshot"), data)

        else:

            if self.audit_enabled:
                self.store.document(self.get_path(record, record_id)).set(data, merge=True)

            if self.snapshot_enabled:
                self.store.document(self.get_path(record_id="snapshot")).set(data)



from django.conf import settings
from firebase_admin import messaging

from bitsoframework.notifications.models import DeviceNotification


class FirebasePushNotificationService(object):

    def send(self, notification: DeviceNotification, **kwargs):
        message = messaging.Message(
            token=notification.device.token, data=notification.data,
            notification=messaging.Notification(title=notification.title,
                                                body=notification.description)
        )

        return messaging.send(message, **kwargs)


class FirebaseWebPushNotificationService(object):

    def __init__(self, icon=None):
        self.icon = icon or getattr(settings, "BITSO_NOTIFICATION_FIREBASE_WEB_PUSH_ICON", None)

    def send(self, notification: DeviceNotification, **kwargs):
        message = messaging.Message(
            token=notification.device.token,
            webpush=messaging.WebpushConfig(
                notification=messaging.WebpushNotification(
                    title=notification.title,
                    body=notification.description,
                    icon=self.icon
                )
            )
        )

        return messaging.send(message, **kwargs)



from bitsoframework.media.filterset import get_filterset_base
from bitsoframework.utils.reflection import require
from django.contrib import admin as django_admin


class MediaRegistry(object):
    defaults = {
        "document": {
            "serializer": "bitsoframework.media.document.serializers.DocumentSerializer",
            "service": "bitsoframework.media.document.services.DocumentService",
            "admin": "bitsoframework.media.document.admin.DocumentAdmin"
        },
        "cloud-document": {
            "serializer": "bitsoframework.media.cloud_document.serializers.CloudDocumentSerializer",
            "service": "bitsoframework.media.cloud_document.services.CloudDocumentService",
            "admin": "bitsoframework.media.cloud_document.admin.CloudDocumentAdmin"
        },
        "image": {
            "serializer": "bitsoframework.media.image.serializers.ImageSerializer",
            "service": "bitsoframework.media.image.services.ImageService",
            "admin": "bitsoframework.media.image.admin.ImageAdmin"
        }
    }

    def __init__(self, storage=[]):
        self.storage = {}

        for entry in storage:
            self.add(**entry)

    @property
    def media_types(self):
        return self.storage.keys()

    def add(self, model, serializer=None, service=None, admin=None, filterset=None, **kwargs):
        model = require(model)

        default = self.defaults.get(model.type, {})

        serializer = serializer or default.get("serializer")
        service = service or default.get("service")
        admin = admin or default.get("admin")
        filterset = filterset or default.get("filterset")

        # if model and admin:
        #    django_admin.site.register(model, admin)

        self.storage[model.type] = {
            "model": model,
            "serializer": serializer,
            "service": service,
            "admin": admin,
            "filterset": filterset,
            **kwargs
        }

    def get_model(self, media_type):
        return self.storage.get(media_type, {}).get("model")

    def get_admin(self, media_type):
        entry = self.storage.get(media_type, {})
        admin = entry.get("admin")

        if isinstance(admin, str):
            admin = require(admin)
            entry["admin"] = admin

        return admin

    def get_serializer(self, media_type):
        entry = self.storage.get(media_type, {})
        serializer = entry.get("serializer")

        if isinstance(serializer, str):
            serializer = require(serializer)
            entry["serializer"] = serializer

        return serializer

    def get_filterset(self, media_type):
        entry = self.storage.get(media_type, {})
        filterset = entry.get("filterset")

        if isinstance(filterset, str):
            filterset = require(filterset)
            entry["filterset"] = filterset

        elif not filterset:
            filterset = get_filterset_base(self.get_model(media_type))
            entry["filterset"] = filterset

        return filterset

    def get_service(self, media_type):
        entry = self.storage.get(media_type, {})
        service = entry.get("service")

        if isinstance(service, str):
            service = require(service)
            entry["service"] = service

        return service

    def register_admin(self, site):
        for media_type in self.storage.keys():
            model = self.get_model(media_type)
            admin = self.get_admin(media_type)
            if admin:
                site.register(model, admin)



from bitsoframework.media.utils import Base64, download

from rest_framework import fields
from rest_framework.fields import CharField
from rest_framework.serializers import Serializer, ModelSerializer

from bitsoframework.media.settings import registry
from bitsoframework.media.utils import Base64, download
from bitsoframework.taggable.mixins import TaggableSerializer


class AbstractMediaSerializer(ModelSerializer):
    uid = fields.ReadOnlyField()

    urls = fields.DictField(read_only=True)

    type = fields.CharField(required=False)

    category = fields.ReadOnlyField()

    # created_by = UserNameSerializer(read_only=True)
    # modified_by = UserNameSerializer(read_only=True)
    created_by = fields.ReadOnlyField()
    modified_by = fields.ReadOnlyField()

    created = fields.ReadOnlyField()
    modified = fields.ReadOnlyField()


class AbstractFileMediaSerializer(AbstractMediaSerializer):
    """
    Serializer for bitso.media.models.Document or bitso.media.models.Image
    """

    extension = fields.ReadOnlyField()

    filesize = fields.ReadOnlyField()

    filename = fields.ReadOnlyField()

    thumbnails_filesize = fields.ReadOnlyField()

    base64 = fields.CharField(write_only=True)
    source_Url = fields.CharField(write_only=True)

    def to_internal_value(self, data):
        base64 = data.pop("base64", None)
        source_url = data.pop("base64", None)
        result = super(AbstractFileMediaSerializer, self).to_internal_value(data)
        if base64:
            file, extension = Base64.get_file(base64)
            result["file"] = file
        if source_url:
            file = download(source_url)
            result["file"] = file

        return result


class MediaMetadataSerializer(TaggableSerializer):
    title = CharField(max_length=255, required=False)
    description = CharField(required=False)


class MediaSerializer(Serializer):
    """
    Helper serializer that emulates support for polymorphic data serialization
    for multiple different types of media model classes.
    """

    def to_representation(self, item):
        serializer_class = registry.get_serializer(item.type)
        serializer = serializer_class()

        return serializer.to_representation(item)


class DictionaryMediaSerializer(Serializer):
    """
    Serializer that renders a list of documents as dictionary where each 
    document's name is a key in the rendered map.
    """

    many = True

    serializer_class = MediaSerializer

    def __init__(self, keys=None, other_key="other", **kwargs):

        super(DictionaryMediaSerializer, self).__init__(**kwargs)

        self.other_key = other_key
        self.keys = keys

    def to_representation(self, items):
        """
        Object instance -> Dict of primitive datatypes.
        """
        map = {}

        other = []

        if self.keys:

            map[self.other_key] = other

            for key in self.keys:
                map[key] = None

        serializer = self.serializer_class()

        for item in items:

            value = serializer.to_representation(item)

            if self.keys:

                if item.name in self.keys:

                    map[item.name] = value

                else:

                    other.append(value)
            else:

                map[item.name] = value

        return map



from threading import local

_active = local()

def get_params(request=None):
    request = request or get_request()

    if hasattr(request, 'data') and request.data:
        return request.data

    return request.POST or request.GET


def get_param(name, request=None, default=None):
    request = request or get_request()

    if name in request.GET:
        return request.GET.get(name, default)

    if name in request.POST:
        return request.POST.get(name, default)

    if hasattr(request, 'data') and request.data and name in request.data:
        return request.data.get(name, default)

    return default


def get_request():
    """
    Utility that gets the current request
    """

    if hasattr(_active, "request"):

        return _active.request
    else:
        return None


def get_user():
    """
    Utility that safely gets the current logged user (when using
    GlobalRequestMiddleware).
    """
    request = get_request()

    if request and request.user.is_authenticated:
        return request.user

    return None


def get_agent():
    """
    Utility that safely gets the current requestor's user agent (when using
    GlobalRequestMiddleware).
    """
    request = get_request()

    if request and request.META:
        return request.META.get("HTTP_USER_AGENT")

    return None


class GlobalRequestMiddleware(object):
    """
    Midlware that attaches the current request into the local thread and makes
    it available via #get_request() function.

    MIDDLEWARE_CLASSES = (
        ...
        'bitso.middleware.GlobalRequestMiddleware'
        ...
    )

    @author: bitsoframework
    """

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        response = self.get_response(request)

        return response

    def process_view(self, request, view_func, view_args, view_kwargs):
        _active.request = request
        return None



import json

from bitsoframework.utils.reflection import require

from bitsoframework.notifications import settings
from bitsoframework.notifications.models import DeviceNotification


class NotificationService(object):
    enabled = settings.NOTIFICATION_ENABLED
    debug = settings.NOTIFICATION_DEBUG

    def __init__(self):
        self.engines = {}

    def get_engine(self, service):
        if service not in self.engines:
            class_name = settings.NOTIFICATION_SERVICES.get(service)
            factory = require(class_name)

            self.engines[service] = factory()

        return self.engines.get(service)

    def push(self, device, recipient, title, description=None, data=None, **kwargs):
        notification = DeviceNotification(device=device,
                                          recipient=recipient,
                                          sender=device.user,
                                          title=title,
                                          description=description,
                                          data=data or {})

        if self.debug:
            print("Pushing notification: " + str(notification))

        if self.enabled:

            engine = self.get_engine(device.service)

            engine.send(notification, **kwargs)

        elif self.debug:
            print("NotificationService is currently disabled.")
            print("To enable, make sure to set BITSO_NOTIFICATION_ENABLE='yes' on your .env file.")

        notification.save()

        return notification



from django.contrib.auth import get_user_model
from django.contrib.auth.tokens import default_token_generator
from django.utils.encoding import force_bytes
from django.utils.http import urlsafe_base64_encode, urlsafe_base64_decode
from rest_framework import serializers, fields


class UpdatePasswordSerializer(serializers.Serializer):
    current_password = serializers.CharField(required=True, max_length=128)
    new_password = serializers.CharField(required=True, max_length=128)

    def validate_current_password(self, current_password):
        user = self.context.get("request").user

        if not user.check_password(current_password):
            self.fail("invalid")

        return current_password

    def update_password(self):
        user = self.context.get("request").user
        user.set_password(self.validated_data.get("new_password"))
        user.save(update_fields=["password"])


class AbstractForgotPasswordSerializer(serializers.Serializer):
    email = serializers.EmailField(required=True)

    def get_token(self, user):
        uid = urlsafe_base64_encode(force_bytes(user.pk))
        token = default_token_generator.make_token(user)
        token_base64 = urlsafe_base64_encode(force_bytes("%s:%s" % (uid, token)))
        return token_base64


class ResetPasswordSerializer(serializers.ModelSerializer):
    password = fields.CharField(required=True)
    token = fields.CharField(required=True)

    class Meta:
        model = get_user_model()
        fields = ('password', 'token')

    def validate_token(self, base64_token):

        try:
            tokens = urlsafe_base64_decode(base64_token).decode().split(":")

            pk = urlsafe_base64_decode(tokens[0]).decode()
            token = tokens[1]

            user = get_user_model().objects.filter(pk=pk).get()

            if not default_token_generator.check_token(user, token):
                self.fail("invalid")

            return user

        except Exception as e:
            self.fail("invalid")

    def reset_password(self):

        user = self.validated_data.get("token")
        password = self.validated_data.get("password")

        user.set_password(raw_password=password)
        user.save(update_fields=['password'])



from django.contrib import admin

from bitsoframework.media.settings import Document


class DocumentAdmin(admin.ModelAdmin):
    model = Document
    list_display = ["id", "title", "filename", "file", "parent_type", "parent_id"]
    search_fields = ["title", "filename"]



from django.db import models

from django.conf import settings
from bitsoframework.media.models import AbstractMedia

ORIGIN_TYPES = getattr(settings, "BITSO_MEDIA_CLOUD_DOCUMENT_ORIGIN_TYPES", {
    "google.com": {
        "document": "google/document",
        "spreadsheets": "google/spreadsheet",
        "presentation": "google/presentation"
    }
})


class AbstractCloudDocument(AbstractMedia):
    """
    Model used to map an online document residing on a external server. Note that sometimes the document
    may translate to a service rather than a file.

    @since: 12/29/2020 1:30:00

    @author: bitsoframework
    """

    type = "cloud-document"
    """
    The unique type of document among the various registered document types.
    """

    origin_type = models.CharField(null=True, blank=True, max_length=255, db_index=True)
    """
    The type of origin for this document, (google doc, google spreadsheet, google presentation, dropbox, etc.).
    """

    class Meta:
        abstract = True

    @property
    def url(self):
        return self.origin_url

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        if not self.origin_type:
            self.origin_type = self.extract_origin_type(self.url)

        return super(AbstractCloudDocument, self).save(force_insert=force_insert, force_update=force_update,
                                                       using=using,
                                                       update_fields=update_fields)

    def extract_origin_type(self, url):

        if not url:
            return None

        for url_fragment, fragments in ORIGIN_TYPES.items():
            if url_fragment in url:
                for fragment, origin_type in fragments.items():
                    if fragment in url:
                        return origin_type



from bitsoframework.utils import files


def calculate(media):
    if hasattr(media, 'filename'):
        return files.get_content_type(media.filename)

    if media.origin_url:
        return files.get_remote_content_type(media.origin_url)

    return None


def get_file(media):
    if isinstance(media, str):
        return open(media, "rb")

    if hasattr(media, "get_file"):
        return media.get_file()

    if hasattr(media, "file"):
        return media.file

    return None



import logging
from threading import Thread

from django.conf import settings
from django.core.mail.message import EmailMultiAlternatives

from bitsoframework.utils import exceptions, template as templates

logger = logging.getLogger(__name__)


def send_as_template(subject, template, recipient_list, context=None,
                     bcc_list=None, cc_list=None, attachments=None,
                     from_email=None, fail_silently=False, connection=None,
                     asynchronous=getattr(settings, "BITSO_ASYNC_EMAIL", False),
                     whitelist=getattr(settings, "BITSO_WHITELISTED_EMAIL_DOMAINS", None),
                     plain_text=False, reply_to=None):
    message = templates.parse(template, context)

    if template.endswith(".html") and not plain_text:

        return send(subject=subject, message=None, from_email=from_email,
                    bcc_list=bcc_list, cc_list=cc_list, attachments=attachments,
                    recipient_list=recipient_list, html_message=message,
                    fail_silently=fail_silently, connection=connection,
                    asynchronous=asynchronous, whitelist=whitelist, reply_to=reply_to)

    else:

        return send(subject=subject, message=message, from_email=from_email,
                    bcc_list=bcc_list, cc_list=cc_list, attachments=attachments,
                    recipient_list=recipient_list, html_message=None,
                    fail_silently=fail_silently, connection=connection,
                    asynchronous=asynchronous, whitelist=whitelist, reply_to=reply_to)


def send(subject, message, recipient_list, html_message=None,
         bcc_list=None, cc_list=None, attachments=None,
         from_email=None, fail_silently=False, connection=None,
         asynchronous=getattr(settings, "BITSO_ASYNC_EMAIL", False),
         whitelist=getattr(settings, "BITSO_WHITELISTED_EMAIL_DOMAINS", None),
         reply_to=None):
    mail = EmailMultiAlternatives(subject=subject, body=message,
                                  from_email=from_email, to=recipient_list, bcc=bcc_list,
                                  connection=connection, attachments=attachments,
                                  cc=cc_list, reply_to=reply_to)

    if whitelist:

        allowed = False

        for domain in whitelist:
            for email in recipient_list:
                if email.endswith(domain):
                    allowed = True
                    break

        if not allowed:
            print("================== EMAIL WHITELIST WARNING ==================")
            print("You're about to send an e-mail to a non-whitelisted domain: " + recipient_list.join(", "))
            print("Here is a JSON document representing the message being sent:")
            print("Subject: " + str(subject))
            print("Message: " + str(message))
            print("HTML Message: " + str(html_message))
            print("=============================================================")

            return

    if html_message:
        mail.attach_alternative(html_message, 'text/html')
        mail.content_subtype = "text/html"

    if asynchronous:

        service = getattr(settings, "BITSO_EMAIL_SERVICE", None)

        if not service:

            service = EmailService()
            service.setDaemon(True)

            settings.BITSO_EMAIL_SERVICE = service

            service.queue(mail)

            service.start()

        else:

            service.queue(mail)

    else:

        try:

            logger.info("Sending message: " + str(mail))

            mail.send()

            logger.info("Message sent")
        except Exception as e:

            logger.exception("Could not send e-mail due:" + exceptions.to_string(e))

            if not fail_silently:
                raise

    return mail


class EmailService(Thread):
    def __init__(self, group=None, target=None, name=None, args=(), kwargs=None, verbose=None):

        Thread.__init__(self, group=group, target=target, name=name, args=args, kwargs=kwargs, verbose=verbose)

        self.messages = []

    def queue(self, message):

        self.messages.append(message)

    def run(self):

        Thread.run(self)

        while len(self.messages) > 0:

            message = self.messages.pop()

            try:

                logger.info("Sending message: " + str(message))

                message.send()

                logger.info("Message sent")


            except Exception as e:
                logger.exception("Could not send e-mail due:" + exceptions.to_string(e))

        settings.BITSO_EMAIL_SERVICE = None



from django.contrib.contenttypes.models import ContentType

from bitsoframework.media import utils
from bitsoframework.utils.models import set_attrs, new_instance


class MediaService(object):
    """
    Service used to manage the lifecycle of the media models.
    """

    category = 0
    """
    The category in which the image is saved against
    """

    model_class = None
    """
    The underlying model class being managed by this service.
    """

    parent_class = None
    """
    The parent related class we are contributing to.
    """

    parent = None
    """
    The parent model related to the maintained images
    """

    def __init__(self, category=0, parent=None, parent_class=None, **kwargs):

        super(MediaService, self).__init__(**kwargs)

        self.category = category
        self.parent = parent
        self.parent_class = parent_class

        if not self.parent_class and self.parent:
            self.parent_class = self.parent.__class__

    def __iter__(self):
        return self.filter().__iter__()

    def get_parent_type(self):

        if not hasattr(self, "_parent_type"):
            self._parent_type = ContentType.objects.get_for_model(self.parent_class)

        return self._parent_type

    def delete(self, **kwargs):

        self.filter(**kwargs).delete()

    def all(self):
        return self.filter()

    def filter(self, **kwargs):

        if self.category:
            kwargs["category"] = self.category

        if self.parent_class:
            kwargs["parent_type"] = self.get_parent_type()

        if self.parent and self.parent.id:
            kwargs["parent_id"] = self.parent.id

        return self.model_class.objects.filter(**kwargs).indexed()  # .select_related("created_by", "modified_by")F

    def download_and_create(self, origin_url, origin_id=None, category=None, **kwargs):

        file = utils.download(origin_url)

        return self.create(file=file, origin_url=origin_url, origin_id=origin_id, category=category, **kwargs)

    def create(self, category=None, auto_save=True, **kwargs):

        model = new_instance(self.model_class,
                             category=category or self.category,
                             **kwargs)

        if self.parent_class:
            model.parent_type = self.get_parent_type()

        if self.parent and self.parent.id:
            model.parent_id = self.parent.id

        if auto_save:
            model.save()

        return model

    def update(self, model, auto_save=True, **kwargs):

        if kwargs.get("tags"):
            model.tags.set(*kwargs.pop("tags"), clear=True)

        set_attrs(model, kwargs)

        if auto_save:
            model.save()

        return model



import math
import random as pr

EMPTY = ""
"""
The empty String <code>""</code>.
"""

INDEX_NOT_FOUND = -1
"""
Represents a failed index search.
"""


def to_string(value, encoding="utf-8"):
    if isinstance(value, bytes):
        return value.decode(encoding)

    if isinstance(value, str):
        return value

    if value is None:
        return value

    return str(value)


def is_empty(string):
    """
    Checks if a String is empty ("") or null.

    @param string: the String to check, may be null

    @return: true if the String is empty or null, false otherwise
    """

    return string is None or len(string) == 0


def tokenize_by_camel_case(string, token):
    """
    Tokenize the given String by camel case replacing upper case characters by
    the given token.
    """

    if not string:
        return string

    result = [string[0:1]]

    index = 1
    length = len(string)

    while index < length:

        character = string[index].upper()

        if character != " " and character == string[index]:

            result.append(token);
            result.append(string[index].lower())

        else:

            result.append(string[index])

        index += 1

    return "".join(result)


def uncapitalize(string):
    """
    UN-Capitalize the first character of the given String
    """

    return string[0].lower() + string[1:]


def contains_ignore_case(string, searchStr):
    """
    <p>
    Checks if String contains a search String irrespective of case, handling
    <code>null</code>. Case-insensitivity is defined as by
    {@link String#equalsIgnoreCase(String)}.

    <p>
    A <code>null</code> String will return <code>false</code>.
    </p>

    @param string
               the String to check, may be null
    @param searchStr
               the String to find, may be null
    @return true if the String contains the search String irrespective of
             case or false if not or <code>null</code> string input
    """

    if (string is None) or (searchStr is None):
        return False

    return string.lower().__contains__(searchStr.lower())


def substring_between(string, before, after):
    """
    <p>
    Gets the String that is nested in between two  Only the first
    match is returned.
    </p>

    <p>
    A <code>null</code> input String returns <code>null</code>. A
    <code>null</code> before/after returns <code>null</code> (no match). An
    empty ("") before and after returns an empty string.
    </p>

    @param string
               the String containing the substring, may be null
    @param before
               the String before the substring, may be null
    @param after
               the String after the substring, may be null
    @return the substring, <code>null</code> if no match
    """

    if (string is None) or (before is None) or (after is None):
        return None

    start = string.find(before);

    if start != INDEX_NOT_FOUND:

        end = string.find(after, start + len(before))

        if end != INDEX_NOT_FOUND:
            return string[start + len(before):end]

    return None


def substring_after(string, delimiter):
    """
    <p>
    Gets the substring after the first occurrence of a delimiter. The
    delimiter is not returned.
    </p>

    <p>
    A <code>null</code> string input will return <code>null</code>. An empty
    ("") string input will return the empty string. A <code>null</code>
    delimiter will return the empty string if the input string is not
    <code>null</code>.
    </p>

    <p>
    If nothing is found, the empty string is returned.
    </p>

    @param string
               the String to get a substring from, may be null
    @param delimiter
               the String to search for, may be null
    @return the substring after the first occurrence of the delimiter, null
            if null String input
    """

    if is_empty(string):
        return string

    if delimiter is None:
        return EMPTY

    pos = string.find(delimiter)

    if pos == INDEX_NOT_FOUND:
        return EMPTY

    return string[pos + delimiter.length()]


def substring_before_last(string, delimiter):
    """
    Gets the substring before the last occurrence of a separator. The
    separator is not returned.
    </p>

    <p>
    A <code>null</code> string input will return <code>null</code>. An empty
    ("") string input will return the empty string. An empty or
    <code>null</code> separator will return the input string.
    </p>

    <p>
    If nothing is found, the string input is returned.
    </p>

    @param string
               the String to get a substring from, may be null
    @param delimiter
               the String to search for, may be null
    @return the substring after the first occurrence of the delimiter, null
            if null String input
    """

    if is_empty(string):
        return string

    if delimiter is None:
        return EMPTY

    pos = string.rfind(delimiter)

    if pos == INDEX_NOT_FOUND:
        return EMPTY

    return string[0:pos]


def split_by_length(value, length, add_suffix=False, suffix_template=" (%i/%i)"):
    """
    Split the given string into an array of substrings at a maximum length.

    :param value:
    :param length:
    :param suffix:
    :return: split array of strings
    """

    max_length = len(value)

    chunks = range(0, int(math.ceil(max_length / length)))
    chunk_size = len(chunks)
    result = []

    for chunk in chunks:
        start = chunk * length
        end = min(max_length, (chunk + 1) * length)
        substring = value[start:end]

        if add_suffix and chunk_size > 1:
            suffix = suffix_template % (chunk + 1, chunk_size)
            substring += suffix

        result.append(substring)

    return result


def random(length=7, validator=None, combination='abcdefghijklmnpqrstuvwxyz0123456789'):
    while True:

        random_id = ''.join(pr.choice(combination) for _ in range(length))

        if not validator or validator(random_id):
            return random_id


class StringBuffer(object):

    def __init__(self, source=None):

        if isinstance(source, (str,)):
            self.source = [source]
        else:
            self.source = source or []

    @property
    def length(self):

        length = 0

        for item in self.source:
            length += len(item)

        return length

    def append(self, value):

        if value:
            self.source.append(value)

    def __str__(self):

        return "".join(self.source)



class ModelPropertyWatcher(object):
    """
    Component responsible for extracting changes off of model instances.
    """

    def __init__(self, include_diff=True, **kwargs):

        self.include_diff = include_diff
        self.properties = kwargs

    def __call__(self, diff):

        for name, handler in self.properties.items():

            old_value = None
            new_value = None

            if name in diff.added():

                new_value = getattr(diff.current, name)

            elif name in diff.removed():

                old_value = getattr(diff.previous, name)

            elif name in diff.changed():

                old_value = getattr(diff.previous, name)
                new_value = getattr(diff.current, name)

            if old_value != new_value:

                if self.include_diff:

                    handler(diff=diff, old_value=old_value, new_value=new_value)

                else:

                    handler(old_value=old_value, new_value=new_value)



from bitsoframework.media.serializers import AbstractMediaSerializer
from bitsoframework.media.settings import CloudDocument


class CloudDocumentSerializer(AbstractMediaSerializer):
    class Meta:
        model = CloudDocument
        exclude = ["parent_type", "parent_id"]



import re


def validate(cnpj):
    """
    Method to validate brazilian cnpjs
    Tests:

    >>> validate('61882613000194')
    True
    >>> validate('61882613000195')
    False
    >>> validate('53.612.734/0001-98')
    True
    >>> validate('69.435.154/0001-02')
    True
    >>> validate('69.435.154/0001-01')
    False
    """
    # defining some variables
    lista_validacao_um = [5, 4, 3, 2, 9, 8, 7, 6, 5, 4, 3, 2]
    lista_validacao_dois = [6, 5, 4, 3, 2, 9, 8, 7, 6, 5, 4, 3, 2]

    # cleaning the cnpj
    cnpj = clean(cnpj)

    # finding out the digits
    verificadores = cnpj[-2:]

    # verifying the lenght of the cnpj
    if len(cnpj) != 14:
        return False

    # calculating the first digit
    soma = 0
    id = 0
    for numero in cnpj:

        # to do not raise indexerrors
        try:
            lista_validacao_um[id]
        except:
            break

        soma += int(numero) * int(lista_validacao_um[id])
        id += 1

    soma = soma % 11
    if soma < 2:
        digito_um = 0
    else:
        digito_um = 11 - soma

    digito_um = str(digito_um)  # converting to string, for later comparison

    # calculating the second digit
    # suming the two lists
    soma = 0
    id = 0

    # suming the two lists
    for numero in cnpj:

        # to do not raise indexerrors
        try:
            lista_validacao_dois[id]
        except:
            break

        soma += int(numero) * int(lista_validacao_dois[id])
        id += 1

    # defining the digit
    soma = soma % 11
    if soma < 2:
        digito_dois = 0
    else:
        digito_dois = 11 - soma

    digito_dois = str(digito_dois)

    # returnig
    return bool(verificadores == digito_um + digito_dois)


def clean(cnpj):
    """
    Method to remove all non numeric values from the given cnpj.
    Tests:

    >>> print clean('53.612.734/0001-98')
    53612734000198
    """
    if cnpj:
        return re.sub('[^0-9]', '', cnpj)
    return cnpj


def format(cnpj):
    """
    Method to format cnpj numbers.
    Tests:

    >>> print format('53612734000198')
    53.612.734/0001-98
    """
    return "%s.%s.%s/%s-%s" % (cnpj[0:2], cnpj[2:5], cnpj[5:8], cnpj[8:12], cnpj[12:14])



from rest_framework.fields import IntegerField, CharField
from rest_framework.serializers import Serializer

from bitsoframework.middleware import get_user


class UserNameSerializer(Serializer):
    """
    Serializer used to expose minimum information about the user (first name,
    last name and id).
    """

    id = IntegerField()

    first_name = CharField()

    last_name = CharField()


class AuthoredSerializer(Serializer):
    def validate(self, value):
        value = super(AuthoredSerializer, self).validate(value)

        if not self.instance:
            value["created_by"] = get_user()

        value["modified_by"] = get_user()

        return value



from django.template.context import Context, RequestContext
from django.template.loader import render_to_string
from django.template.base import Template
from bitsoframework.middleware import get_request


def parse(template, context=None):
    """
    Parse the given template String or template name using the given context dictionary
    and return its result as a String

    @param template: the template name (if ending with *.html) or the template
    String itself

    @param context the dictionary holding additional variables

    @return: parsed template content
    """
    context = context or {}

    if template.endswith(".html") or template.endswith(".text"):

        return render_to_string(template, context=context)

    else:

        request = get_request()
        context = RequestContext(request, context) if request else Context(context)

        template = Template(template)

        return template.render(context)



from django.db import models
from django.db.models.base import Model

from bitsoframework.media.image.services import ImageService


class MyProduct(Model):
    name = models.CharField(max_length=50)

    def __init__(self, *args, **kwargs):
        Model.__init__(self, *args, **kwargs)

        self.images = ImageService(parent=self, category=0)
        self.defects = ImageService(parent=self, category=1)



import os
from tempfile import mkdtemp

from django.conf import settings
from django.core.files import File
from django.core.files.storage import FileSystemStorage
from django.test import TestCase

from bitsoframework.media.settings import Image
from .models import MyProduct


class ImageServiceTestCase(TestCase):
    def setUp(self):
        image_field = Image()._meta.get_field_by_name('file')[0]

        self.product = MyProduct.objects.create(name="Macbook Pro 15\"")

        image_field.storage = FileSystemStorage(location=mkdtemp(), base_url="/")

    def test_crud(self):
        """Make sure images are attached, retrieved and deleted"""

        f = File(open(settings.BASE_DIR + "/bitsoframework/tmp/Photo-1.jpg"))

        photo = self.product.images.save(name="Photo 1", file=f)
        photo_path = photo.file.file.name

        self.assertTrue(os.path.exists(photo_path), "Photo does not exist on local path")
        self.assertEqual(photo.width, 350.0, "Photo 1's width does not match")
        self.assertEqual(photo.height, 277.0, "Photo 1's height does not match")

        self.product = MyProduct.objects.get(id=self.product.id)

        self.assertEqual(self.product.images.filter().count(), 1, "Photo 1 is not created")

        photo = self.product.images.filter()[0]

        self.assertEqual(photo.width, 350.0, "Photo 1's width does not match")
        self.assertEqual(photo.height, 277.0, "Photo 1's height does not match")

        photo.file = File(open(settings.BASE_DIR + "/bitsoframework/tmp/Photo-2.jpg"))
        photo.save()

        self.assertTrue(os.path.exists(photo_path), "Photo should exist on same local path")
        self.assertEqual(self.product.images.filter().count(), 1, "Photo should be updated...")

        self.assertEqual(photo.width, 275.0, "Photo 2's width does not match")
        self.assertEqual(photo.height, 183.0, "Photo 2's height does not match")

        photo.delete()

        self.assertFalse(os.path.exists(photo_path), "Photo should not exist on local path")
        self.assertEqual(self.product.images.filter().count(), 0, "Photo should be deleted...")



from django.conf import settings
from django.db.models.signals import pre_delete
from django.dispatch.dispatcher import receiver

from bitsoframework.media.registry import MediaRegistry
from bitsoframework.utils.reflection import require

registry = MediaRegistry(getattr(settings, "BITSO_MEDIA_REGISTRY", []))

Document = registry.get_model("document")
Image = registry.get_model("image")
CloudDocument = registry.get_model("cloud-document")

PREVIEW_ENABLED = getattr(settings, "BITSO_DOCUMENT_PREVIEW_ENABLED", True)

PREVIEW_WIDTH = getattr(settings, "BITSO_DOCUMENT_PREVIEW_WIDTH", 768)
PREVIEW_HEIGHT = getattr(settings, "BITSO_DOCUMENT_PREVIEW_HEIGHT", 1024)

AVOID_DUPLICATES = getattr(settings, "BITSO_MEDIA_AVOID_DUPLICATES", True)

CHECKSUM_CALCULATOR = require(
    getattr(settings, "BITSO_MEDIA_CHECKSUM_CALCULATOR", "bitsoframework.media.checksum.calculate_sha256"))

CONTENT_TYPE_CALCULATOR = require(
    getattr(settings, "BITSO_MEDIA_CONTENT_TYPE_CALCULATOR", "bitsoframework.media.content_type.calculate"))

if Image:
    @receiver(pre_delete, sender=Image)
    def on_delete_image(sender, instance, **kwargs):
        instance.delete_file(instance.file)

if Document:
    @receiver(pre_delete, sender=Document)
    def on_delete_document(sender, instance, **kwargs):
        instance.delete_file(instance.file)



from rest_framework.exceptions import ValidationError
from rest_framework.fields import Field, FloatField


class CoordinatesField(Field):
    latitude = FloatField()
    longitude = FloatField()

    def to_internal_value(self, data):

        if data is not None:
            if not isinstance(data, dict):
                raise ValidationError("Expecting {latitude:..., longitude:...} but got %s" % (str(data),))

            latitude = data.get("latitude")
            longitude = data.get("longitude")

            latitude = self.latitude.to_internal_value(latitude)
            longitude = self.longitude.to_internal_value(longitude)

            from django.contrib.gis.geos.point import Point

            return Point(x=longitude,
                         y=latitude)

    def to_representation(self, instance):
        if instance:
            return {
                "longitude": instance.x,
                "latitude": instance.y
            }



from bitsoframework.utils import lists


def get_data(model, excluded=[]):
    data = {}

    for fld in model._meta.fields:
        if len(excluded) > 0 and excluded.__contains__(fld.name):
            continue

        data[fld.name] = getattr(model, fld.name)

    return data


def new_instance(model_class, **kwargs):
    data = {}

    for fld in model_class._meta.fields:
        if fld.name not in kwargs:
            continue

        data[fld.name] = kwargs.get(fld.name)

    model = model_class(**data)

    return model


def clone(model):
    new_kwargs = get_data(model, excluded=[model._meta.pk])

    return get_model_class(model)(**new_kwargs)


def get_model_class(model):
    return model.__class__


def get_field(clazz, name):
    for field in clazz._meta.fields:
        if field.name == name:
            return field

    return None


def set_attrs(object, __map=None, __allow_null=True, **kwargs):
    if not __map:
        __map = kwargs
    else:
        kwargs.update(__map)

    for key, value in __map.items():

        if not __allow_null and value is None:
            continue

        setattr(object, key, value)

    return object


def get_choice_label(choices, choice):
    for entry in choices:

        if entry[0] == choice:
            return entry[1]

    return ""


def get_choice_id(tuples, str_value):
    if str_value:
        s = str_value.lower()
        for t in tuples:
            if t[1].lower() == s:
                return t[0]
    return None


def exists(record):
    clazz = get_model_class(record)
    pk = get_field(clazz, 'id')

    if pk.auto_created:
        return record.pk is not None

    return clazz.objects.filter(pk=record.pk).exists()


def diff_list(old_value, new_value):
    old_value = lists.as_list(old_value)
    new_value = lists.as_list(new_value)

    added = []
    removed = []

    for item in new_value:
        if item not in old_value:
            added.append(item)

    for item in old_value:
        if item not in new_value:
            removed.append(new_value)

    return {
        "added": added,
        "removed": removed
    }



from django.db.models import SET_NULL, ForeignKey, Model
from rest_framework import status
from rest_framework.decorators import action
from rest_framework.exceptions import ValidationError
from rest_framework.parsers import FormParser, MultiPartParser, FileUploadParser
from rest_framework.response import Response

from bitsoframework.media.image.serializers import ImageSerializer
from bitsoframework.media.image.services import ImageService
from bitsoframework.media.serializers import MediaMetadataSerializer
from bitsoframework.media.settings import Image
from bitsoframework.media.utils import is_image


class PhotoModelMixin(Model):
    photo = ForeignKey(Image, null=True, blank="true", on_delete=SET_NULL)

    class Meta:
        abstract = True


class AbstractPhotoViewMixin(object):
    delete_existing_photo = True

    def get_photo_for(self, request, instance):

        serializer = ImageSerializer(instance=instance.photo)

        return Response(serializer.data)

    def create_photo_for(self, request, instance):
        """
        Attach a new photo to the underlying record.
        """

        file = None

        for item in request.data.values():

            if not hasattr(item, "file"):
                continue

            file = item

        if file is None:
            raise ValidationError("A single image file should be uploaded")

        if not is_image(file):
            raise ValidationError("The uploaded file is not a valid image")

        serializer = MediaMetadataSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        service = ImageService()

        params = {}

        if serializer.validated_data:
            params.update(serializer.validated_data)

        if instance.photo and self.delete_existing_photo:
            instance.photo.delete()

        photo = service.create(file=file, **params)

        instance.photo = photo
        instance.save()

        serializer = ImageSerializer(photo)

        return Response(serializer.data, status=status.HTTP_201_CREATED)

    def update_photo_for(self, request, instance):

        service = ImageService()

        serializer = ImageSerializer(data=request.data, partial=True)
        serializer.is_valid(raise_exception=True)

        if instance.photo:
            service.update(instance.photo, **serializer.validated_data)

        serializer = ImageSerializer(instance.photo)

        return Response(serializer.data, status=status.HTTP_200_OK)

    def remove_photo_for(self, request, instance):

        if instance.photo:
            photo = instance.photo

            instance.photo = None
            instance.save()

            photo.delete()

        return Response(status=status.HTTP_200_OK)


class PhotoViewMixin(AbstractPhotoViewMixin):
    @action(detail=False, methods=["get"], url_path="photo")
    def get_photo(self, request):
        """
        Attach a new photo to the underlying record.
        """

        instance = self.get_photo_owner()

        return self.get_photo_for(request, instance)

    @action(detail=False, methods=["put"], url_path="photo/create",
            parser_classes=(FormParser, MultiPartParser, FileUploadParser))
    def create_photo(self, request):
        """
        Attach a new photo to the underlying record.
        """

        instance = self.get_photo_owner()

        return self.create_photo_for(request, instance)

    @action(detail=False, methods=["patch"], url_path="photo/update")
    def update_photo(self, request):
        """
        Update data for an existing photo record. Notice that updating the
        file itself is not supported.
        """

        instance = self.get_photo_owner()

        return self.update_photo_for(request, instance)

    @action(detail=False, methods=["delete"], url_path="photo/destroy")
    def remove_photo(self, request):
        """
        Remove the currently attached photo to the underlying record.
        """

        instance = self.get_photo_owner()

        return self.remove_photo_for(request, instance)


class MyPhotoViewMixin(PhotoViewMixin):

    def get_photo_owner(self):
        return self.request.user


class PhotoDetailViewMixin(AbstractPhotoViewMixin):
    @action(detail=True, methods=["get"], url_path="photo")
    def get_photo(self, request, pk):
        """
        Attach a new photo to the underlying record.
        """

        instance = self.get_object()

        return self.get_photo_for(request, instance)

    @action(detail=True, methods=["put"], url_path="photo/create",
            parser_classes=(FormParser, MultiPartParser, FileUploadParser))
    def create_photo(self, request, pk):
        """
        Attach a new photo to the underlying record.
        """

        instance = self.get_object()

        return self.create_photo_for(request, instance)

    @action(detail=True, methods=["patch"], url_path="photo/update")
    def update_photo(self, request, pk):
        """
        Update data for an existing photo record. Notice that updating the
        file itself is not supported.
        """

        instance = self.get_object()

        return self.update_photo_for(request, instance)

    @action(detail=True, methods=["delete"], url_path="photo/destroy")
    def remove_photo(self, request, pk):
        """
        Remove the currently attached photo to the underlying record.
        """

        instance = self.get_object()

        return self.remove_photo_for(request, instance)



import os

from django.conf import settings
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.files.base import File
from django.db import models
from django.db.models import CASCADE
from django.db.models.fields.files import FieldFile
from django.utils.functional import cached_property
from django_extensions.db.models import TimeStampedModel, TitleDescriptionModel, ActivatorModel

from bitsoframework.authored.models import AuthoredModel
from bitsoframework.indexable.models import IndexedModel
from bitsoframework.media.utils import get_uid
from bitsoframework.utils import files
from bitsoframework.utils.files import get_random_name
from bitsoframework.utils.reflection import require

ON_DELETE = getattr(settings, 'BITSO_MEDIA_ON_DELETE', CASCADE)
MEDIA_CATEGORY_CHOICES = getattr(settings, "BITSO_MEDIA_CATEGORY_CHOICES", None)
MEDIA_PATH = getattr(settings, "BITSO_MEDIA_PATH", "uploads/documents")
MEDIA_PATH_RESOLVER_CLASS = require(getattr(settings, "BITSO_MEDIA_PATH_RESOLVER",
                                            "bitsoframework.media.path_resolvers.MediaPathResolver"))
MEDIA_PATH_RESOLVER = MEDIA_PATH_RESOLVER_CLASS(path=MEDIA_PATH)

DOCUMENT_MODEL_NAME = getattr(settings, "BITSO_MEDIA_DOCUMENT_MODEL_NAME", "media.Document")


class AbstractMedia(IndexedModel,
                    TitleDescriptionModel,
                    ActivatorModel,
                    TimeStampedModel,
                    AuthoredModel):
    """
    Abstract class common to document and image models.
    
    @since: 05/10/2016 19:50:00
    
    @author: bitsoframework
    """

    class Meta:
        abstract = True

    type = "undefined"
    """
    The type of media we are dealing with (overridden by every sub-class).
    """

    parent_id = models.CharField(db_index=True, max_length=200, null=True, blank=True)
    """
    The parent model's identifier.
    """

    parent_type = models.ForeignKey(ContentType, on_delete=ON_DELETE, null=True, blank=True, related_name="%(class)ss")
    """
    The type bound to this document
    """

    parent = GenericForeignKey('parent_type', 'parent_id')
    """
    The actual parent object storing this document.
    """

    category = models.IntegerField(db_index=True, null=True, blank=True, choices=MEDIA_CATEGORY_CHOICES)
    """
    The category used to group the image in the target model.
    """

    origin_url = models.URLField(null=True, blank=True, max_length=1000)
    """
    The URL from where this image was originally imported.
    """

    origin_id = models.CharField(null=True, blank=True, max_length=255, db_index=True)
    """
    An extra bit of information about the origin of this image.
    """

    content_type = models.CharField(max_length=200, null=True, blank=True)
    """
    The type of content mapped by this file.
    """

    def __str__(self):
        return self.title

    @cached_property
    def uid(self):
        return get_uid(self)

    @cached_property
    def service(self):
        from .settings import registry

        service_class = registry.get_service(self.type)

        return service_class(self)

    @property
    def urls(self):
        return {
            "original": self.url
        }

    def calculate_content_type(self):
        from .settings import CONTENT_TYPE_CALCULATOR
        return CONTENT_TYPE_CALCULATOR(self)

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        if not self.content_type:
            self.content_type = self.calculate_content_type()

        super(AbstractMedia, self).save(force_insert=force_insert, force_update=force_update, using=using,
                                        update_fields=update_fields)


class AbstractFileMedia(AbstractMedia):
    filename = models.CharField(max_length=200, null=True, blank=True)
    """
    The original file name if needed to be used for anything
    """

    checksum = models.CharField(max_length=256, null=True, blank=True)
    """
    The original file's checksum
    """

    filesize = models.IntegerField(null=True, blank=True)
    """
    The number of bytes for the original file (size).
    """

    class Meta:
        abstract = True

    def __init__(self, *args, **kwargs):

        super(AbstractFileMedia, self).__init__(*args, **kwargs)

        if self.pk:
            self.old_file = self.file if hasattr(self, "file") else None
        else:
            self.old_file = None

    @property
    def url(self):
        return self.file.url

    @property
    def extension(self):

        if self.filename:
            name, extension = os.path.splitext(self.filename)

            return extension[1:].lower()

        return None

    @cached_property
    def storage(self):
        return self._meta.get_field('file').storage

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):

        if self.file != self.old_file and self.file and (update_fields is None or "file" in update_fields):
            self.on_update_file()

        return super(AbstractFileMedia, self).save(force_insert=force_insert, force_update=force_update, using=using,
                                                   update_fields=update_fields)

    def on_update_file(self):

        self.filename = files.get_filename(self.get_file().name)
        self.filesize = self.get_file().size

        if not self.title:
            self.title = self.filename

        if self.old_file:
            # delete existing file so we can override it
            self.delete_file(self.old_file)

            #
            self.file.name = self.old_file.name
            self.checksum = None
            self.content_type = None

        self.file.name = get_random_name(self.filename)

        if not self.checksum:
            self.checksum = self.calculate_checksum()

        if not self.content_type:
            self.content_type = self.calculate_content_type()

    def get_file(self):

        if isinstance(self.file, FieldFile):
            return self.file.file

        return self.file

    def calculate_checksum(self):
        from .settings import CHECKSUM_CALCULATOR

        return CHECKSUM_CALCULATOR(self)

    def copy(self, **kwargs):

        new_file = File(self.file.storage.open(self.file.name))

        copy = super(AbstractFileMedia, self).copy(file=new_file, old_file=None, **kwargs)

        return copy

    def resolve_path(self, filename, resolver):
        """
        Method called finally resolve the location where this file should be stored.
        At first, the path is resolved by the framework over the bitso's MEDIA_PATH_RESOLVER settings
        but ultimately models have the ability to override this if necessary.

        For more info, see bitso.media.path_resolvers

        @return the resolved filepath for the given filename under this model. If nothing is returned
        then the logic is delegated up to the configured settings instead.
        """
        return None



from django.contrib.auth import get_user_model
from django.contrib.auth.hashers import check_password

from django.contrib.auth.backends import ModelBackend

UserModel = get_user_model()


class EmailBackend(ModelBackend):
    """
    Authenticates against settings.AUTH_USER_MODEL by email.
    """

    def authenticate(self, request, username=None, password=None, **kwargs):
        if username is None:
            username = kwargs.get(UserModel.USERNAME_FIELD)
        try:
            user = UserModel._default_manager.get(email=username)
        except UserModel.DoesNotExist:
            # Run the default password hasher once to reduce the timing
            # difference between an existing and a nonexistent user (#20760).
            UserModel().set_password(password)
        else:
            if user.check_password(password) and self.user_can_authenticate(user):
                return user



from django.conf import settings

NOTIFICATION_SERVICES = getattr(settings, "BITSO_NOTIFICATION_SERVICES", {
    "firebase-push": "bitsoframework.notifications.services.firebase.FirebasePushNotificationService",
    "firebase-web-push": "bitsoframework.notifications.services.firebase.FirebaseWebPushNotificationService"
})

NOTIFICATION_ENABLED = getattr(settings, "BITSO_NOTIFICATION_ENABLED", False)
NOTIFICATION_DEBUG = getattr(settings, "BITSO_NOTIFICATION_DEBUG", False)



import simplejson
from django_extensions.db.models import ActivatorModel
from rest_framework import status
from rest_framework.decorators import action
from rest_framework.exceptions import ValidationError
from rest_framework.parsers import FormParser, MultiPartParser, FileUploadParser
from rest_framework.response import Response
from django_filters import utils as df_utils

from bitsoframework.media import settings
from bitsoframework.media.serializers import *
from bitsoframework.media.utils import is_image, get_files, download


class MediaServiceMixin(object):
    """
    shared functionality for services to create and destroy media objects
    """

    avoid_duplicates = settings.AVOID_DUPLICATES

    def import_media(self, record, origin_url, origin_id=None, category=None, **kwargs):

        file = download(origin_url)

        return self.create_media(record, file=file, origin_url=origin_url, origin_id=origin_id, category=category,
                                 **kwargs)

    def create_media(self, record, file=None, category=None, **kwargs):
        """
        @param record: parent record that the media belongs to

        @param file: file that is being attached as a document/image

        @param category: multiple types of files can be uploaded to a record and
                         when this happens we need to pass in the file category
        """
        type = kwargs.pop("type", None)

        if not type and file:
            if is_image(file):
                type = "image"

        if not type:
            type = "document"

        service_class = registry.get_service(type)
        service = service_class(parent=record)

        if file and self.avoid_duplicates:
            checksum = settings.CHECKSUM_CALCULATOR(file)
            media = service.filter(checksum=checksum, status=ActivatorModel.ACTIVE_STATUS).first()
            if media:
                return media
            else:
                kwargs["checksum"] = checksum

        media = service.create(category=category, file=file, **kwargs)

        return media

    def update_media(self, record, media_id, media_type, **kwargs):
        """
        Update media information such as name or description for the given
        media object within the given record.

        Note that updating media's file is not supported. This can only be done
        by deleting the existing file and adding a new media record instead.
        """

        service = record.media_service_from_type(media_type)

        media = service.filter(id=media_id).get()

        return service.update(media, **kwargs)

    def destroy_media(self, record, media_id, media_type):

        service = record.media_service_from_type(media_type)

        service.delete(id__in=[media_id])


class MediaViewMixin(object):
    list_media_serializer_class = MediaSerializer
    create_media_response_serializer_class = MediaSerializer
    media_service_class = MediaServiceMixin
    """
    shared functionality for views to interact with media objects
    """

    @action(detail=True, url_path="media")
    def list_media(self, request, pk):

        return self.list_media_by_type(request, pk, None)

    @action(detail=True, url_path="media/(?P<media_type>[-\w]+)")
    def list_media_by_type(self, request, pk, media_type):
        """
        Retrieve all media documents attached to the underlying record.

        @param media_type: the type of media (image|document)
        """

        instance = self.get_object()

        media_types = [media_type] if media_type else registry.media_types

        media = []

        for media_type in media_types:

            queryset = instance.media_service_from_type(media_type).all()

            filterset_class = registry.get_filterset(media_type)
            filterset = filterset_class(queryset=queryset, request=request, data=request.query_params)

            if not filterset.is_valid():
                raise df_utils.translate_validation(filterset.errors)

            queryset = filterset.filter_queryset(queryset)

            for item in queryset:
                media.append(item)

        serializer = self.list_media_serializer_class(instance=media, many=True)

        return Response(serializer.data)

    @action(detail=True, methods=["delete"], url_path="media/destroy/(?P<media_id>[\d+]+)/(?P<media_type>[-\w]+)")
    def destroy_media(self, request, pk, media_id, media_type):
        """
        Destroys an existing media entry within the given object, permanently 
        deleting all of its metadata and attached files - including thumbnails 
        for images.
        
        @param media_id: the media's primary key
        
        @param media_type: the type of media (image|document)
        """

        instance = self.get_object()

        service = self.media_service_class()

        service.destroy_media(instance, media_id, media_type)

        return Response(status=status.HTTP_200_OK)

    @action(detail=True, methods=["put"], url_path="media/create",
            parser_classes=(FormParser, MultiPartParser, FileUploadParser))
    def create_media(self, request, pk, *args, **kwargs):
        """
        Attach a new media content to the bounding record.
                    
        @param metadata: An array of objects (JSON), each object containing name
            of the document being uploaded. Note that while sending multiple files,
            both file and metadata should match size and order.
        """

        record = self.get_object()

        service = self.media_service_class()

        documents = MediaViewMixin._create_media(request, service, record)

        serializer = self.create_media_response_serializer_class(instance=documents, many=True)

        return Response(serializer.data, status=201)

    @classmethod
    def _create_media(cls, request, service, record, **kwargs):

        metadata = simplejson.loads(request.data.get("metadata")) if "metadata" in request.data else None
        data = simplejson.loads(request.data.get("data")) if "data" in request.data else None

        documents = list()

        files = get_files(request)

        i = 0
        for file in files:

            params = {
                **kwargs
            }

            if metadata:
                params.update(metadata[i])

            documents.append(service.create_media(record, file=file, **params))

            i += 1

        if data:

            for item in data:
                serializer_class = registry.get_serializer(item.get("type"))
                serializer = serializer_class(data=item)
                serializer.is_valid(raise_exception=True)

                documents.append(service.create_media(record, **serializer.validated_data))

        return documents

    @action(detail=True, methods=["patch"], url_path="media/update/(?P<media_id>[\d+]+)/(?P<media_type>[-\w]+)")
    def update_media(self, request, pk, media_id, media_type, *args, **kwargs):
        """
        Update data for an existing media record. Notice that updating the
        file itself is not supported.
        """

        record = self.get_object()

        service = self.media_service_class()

        serializer_class = registry.get_serializer(media_type)

        files = get_files(request)

        params = {}

        # attach file being updated
        if len(files) == 1:
            params["file"] = files[0]

        data = simplejson.loads(request.data.get("metadata")) if "metadata" in request.data else request.data

        if data:
            serializer = serializer_class(data=data, partial=True)
            serializer.is_valid(raise_exception=True)

            params.update(serializer.validated_data)

        params.update(kwargs)

        media = service.update_media(record, media_id, media_type, **params)

        serializer = serializer_class(instance=media)

        return Response(serializer.data, status=status.HTTP_200_OK)


class MediaModelMixin(object):
    """
    shared functionality for models that manage bitso media objects.
    """

    def media_service_from_type(self, media_type):
        """
        Look up for a media service based on its type
        
        @param media_type: the type of media service
        
        @return service for the given media type (either ImageService or DocumentService)
        """
        from bitsoframework.media.settings import registry

        service_class = registry.get_service(media_type)

        return service_class(parent=self)

    def get_media(self, media_type=None, **kwargs):
        result = []

        if media_type:
            for record in self.media_service_from_type(media_type).filter(**kwargs):
                result.append(record)
        else:

            from bitsoframework.media.settings import registry

            for media_type in registry.media_types:

                for record in self.media_service_from_type(media_type).filter(**kwargs):
                    result.append(record)

        return result

    @property
    def media(self):
        """
        Encapsulate all media documents and images. If there are different categories
        of documents this method will need to be overridden.
        """

        return self.get_media()

    def copy(self, **kwargs):
        """
        Copies all underlying documents under new instances/paths in the remote server.
        :param kwargs:
        :return:
        """

        new_model = super(MediaModelMixin, self).copy(**kwargs)

        for document in self.media:
            document.copy(parent_id=new_model.id)

        return new_model



from bitsoframework.media.document.models import AbstractDocument
from bitsoframework.media.image.models import AbstractImage
from bitsoframework.media.cloud_document.models import AbstractCloudDocument


class Image(AbstractImage):
    pass


class Document(AbstractDocument):
    pass


class CloudDocument(AbstractCloudDocument):
    pass



def get_full_name(user=None, first_name=None, last_name=None):
    if user:
        first_name = user.first_name
        last_name = user.last_name

    if first_name and last_name:
        return first_name + " " + last_name

    if last_name:
        return last_name

    return first_name


def get_first_and_last_name(full_name):
    if not full_name:
        return "", ""

    names = full_name.split(" ")

    length = len(names)

    first_name = names[0]
    last_name = " ".join(names[1:length]) if length > 1 else ""

    return first_name, last_name
